/*-
 * Copyright 2013 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <assert.h>
#include <strings.h>
#include <sys/errno.h>
#include "cx88_debug.h"
#include "cx88_analog_driver_factory.h"

struct cx88_analog_driver_entry
{
   const char *name;
   cx88_analog_driver_factory::constructFn ctor;
};

#define CX88_MAX_ANALOG_DRIVERS 32

static cx88_analog_driver_entry drivers[CX88_MAX_ANALOG_DRIVERS];

static uint32_t num_drivers = 0;

void cx88_analog_driver_factory::registerDriver(const char *name,
                                                cx88_analog_driver_factory::constructFn ctor)
{
   assert(num_drivers < CX88_MAX_ANALOG_DRIVERS);
   drivers[num_drivers].name = name;
   drivers[num_drivers].ctor = ctor;
   ++num_drivers;
}

cx88_analog_driver *cx88_analog_driver_factory::create(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   cx88_analog_driver::cx88_analog_mode mode,
   cx88_analog_driver *slave,
   cx88_analog_driver *pip_driver,
   int &error)
{
   cx88_xml_config::card_desc desc;
   error = (error ? error : cx88_driver::get_card_desc(cx88conf, devnode, desc));
   if (error)
   {
      error = (error ? error : ENOENT);
      return NULL;
   }
   if (desc.analog_class == NULL)
   {
      desc.analog_class = "cx88_analog_driver";
   }
   for (uint32_t i = 0; i < num_drivers; ++i)
   {
      if (strcasecmp(desc.analog_class, drivers[i].name) == 0)
      {
         if (desc.options != NULL)
         {
            error = tunerconf.load_string(desc.options, ';');
         }
         return drivers[i].ctor(desc.name, cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error);
      }
   }
   CX88ERR << "Analog driver class " << desc.analog_class << " not found.  Available classes:" << std::endl;
   for (uint32_t i = 0; i < num_drivers; ++i)
   {
      CX88SUBERR << '\t' << drivers[i].name << std::endl;
   }
   return NULL;
}
