/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <libgen.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/time.h>
#include <glob.h>

#include <new>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <list>
#include <iomanip>
#include <ext/stdio_filebuf.h>

#include "cx88_debug.h"
#include "cx88_capture_factory.h"
#include "cx88_mpeg_driver_factory.h"
#include "cx88_analog_driver_factory.h"

using namespace std;
using namespace __gnu_cxx;

#ifndef DEFAULT_XML_CONFIG
   #define DEFAULT_XML_CONFIG ".cx88.xml"
#endif
#ifndef MAX_TCP_CONNECTIONS 
   #define MAX_TCP_CONNECTIONS 8
#endif

extern char *optarg;
extern int optind;
extern int optreset;

typedef struct
{
   cx88_driver *driver;
   double threshold;
   uint32_t lock_timeout_ms;
} cx88_scan_arg;

typedef struct
{
   cx88_driver *driver;
   capvec captures;
   char *device;
   tuner_config tunerconf;
   cx88_driver::cx88_capture_mode capture_mode;
} cx88_driver_entry;

typedef std::list<cx88_driver_entry> driverlist;

static int capture(
   cx88_driver_entry &entry,
   cx88_xml_config &xmlconf,
   string &channel,
   unsigned long duration_s,
   unsigned short remote_port,
   bool remote_enable,
   bool auto_start,
   bool interactive,
   cx88_capture_params &params);

static void print_channel(void *arg, int error);
   
static void capture_complete(void *arg, int error);

static int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params);

static int push_driver_entry(driverlist &drivers)
{
   try
   {
      cx88_driver_entry entry;
      entry.driver = NULL;
      entry.device = NULL;
      entry.capture_mode = cx88_driver::CX88_CAPTURE_MODE_NORMAL;
      drivers.push_back(entry);
   }
   catch (...)
   {
      CX88ERR << "Unable to allocate memory for driver entry structure" << endl;
      return ENOMEM;  
   }
   return 0;
}

static int g_controlfds[2];

static void term_handler(int)
{
   capture_complete(&(g_controlfds[1]), 0);
}

static void list_devices(cx88_xml_config &xmlconf, const char *header, const char *pattern)
{
   glob_t g;
   glob(pattern, 0, NULL, &g);
   char **paths = g.gl_pathv;
   cout << header << endl;
   ofstream temperr("/dev/null");
   streambuf *origerr = cerr.rdbuf(temperr.rdbuf());
   if (paths != NULL)
   {
      while (*paths != NULL)
      {
         cx88_xml_config::card_desc desc;
         if (cx88_driver::get_card_desc(xmlconf, *paths, desc) == 0)
         {
            cout << '\t' << *paths << ": ";
            if (desc.name == NULL) cout << "Unknown card (id: 0x" << hex << desc.id << dec << ")" << endl; 
            else cout << desc.name << endl;
         }
         ++paths;
      }
   }
   cerr.rdbuf(origerr);
   globfree(&g);
}

int main(int argc, char **argv)
{
   driverlist drivers;
   int error = 0, optch = -1;
   string xml_file = DEFAULT_XML_CONFIG;
   string channel = CX88_XML_CHANNEL_DEFAULT;
   unsigned long duration_s = 0;
   unsigned short remote_port = 0;
   bool remote_enable = false;
   bool auto_start = true;
   bool interactive = true;
   bool switch_audio = false;
   cx88_capture_params params;
   params.completion = NULL;
   params.completion_arg = NULL;
   params.channel_event = NULL;
   params.channel_arg = NULL;
   params.lock_timeout_ms = 1000;
   g_controlfds[0] = -1;
   g_controlfds[1] = -1;
   static const char *help = 
      "Options:\n"
      "\t-d <path>: [Required] Specify device from which to capture\n"
      "\t-c <channel | profile:channel>: [Optional] Specify initial channel, e.g. '7' or 'USA:7'\n"
      "\t-u <protocol://hostname:port>: [Optional] URL, Send capture over TCP or UDP to provided host\n"
      "\t-x <filename>: [Optional] Specify XML configuration file; default is "DEFAULT_XML_CONFIG"\n"
      "\t-t <filename>: [Optional] Specify libtuner configuration file for current device\n"
      "\t-o <string>: [Optional] Specify custom options in the form of <key0>=<value0>;<key1>=<value1>;...\n"
      "\t-n <value>: [Optional] Terminate capture after specified number of minutes\n"
      "\t-p <value>: [Optional] Specify port on which to listen for remote control connections\n"
      "\t-l <value> [Optional] Maximum channel lock timeout, in ms\n"
      "\t-q [Optional] Non-interactive capture--do not listen for input on stdin\n"
      "\t-h [Optional] Do not automatically start capturing\n"
      "\t-a [Optional] Force normal-mode (non-OSS) audio capture even when no capture URLs are specified\n"
      "\t-b [Optional] Backend capture mode.  Disable frontend devices like tuners and demodulators\n"
      "\t-v : Display version\n"
      "\t-e : List available devices\n"
      "\t-? : Display this message\n";
   static const char *opts = "d:abc:u:x:t:n:o:p:l:qhve?";
   // Globally ignore SIGPIPE to keep from dying if the other end of a capture/connection drops out
   signal(SIGPIPE, SIG_IGN);
   if (pipe(g_controlfds))
   {
      CX88ERR << "Unable to create control file descriptor: " 
         << strerror(errno) << endl;  
      error = errno;
   }
   if (!error)
   {
      signal(SIGTERM, term_handler);
      signal(SIGINT, term_handler);
      error = push_driver_entry(drivers);
   }
   bool failquiet = false;
   while ((optch = getopt(argc, argv, opts)) != -1)
   {
      switch (optch)
      {
         case 'x':
            try
            {
               xml_file = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store XML configuration file name.";
               error = (error ? error : ENOMEM);
            }
            break;
         case 'v':
            cout << "cx88 " CX88_VERSION << endl;
            failquiet = true;
            break;
         case '?':
            cout << help;
            failquiet = true;
         default:
            break;
      }
   }
   optind = 1;
   optreset = 1;
   cx88_xml_config xmlconf(xml_file.c_str(), error);
   cx88_capture_factory::init(xmlconf, error);
   while (!error && ((optch = getopt(argc, argv, opts)) != -1))
   {
      switch (optch) 
      {
         case 'd':
         {
            char *device = strdup(optarg);
            if (device == NULL)
            {
               CX88ERR << "Unable to store device filename." << endl;
               error = ENOMEM;
            }
            if (drivers.back().device != NULL)
            {
               error = push_driver_entry(drivers);
            }
            if (!error)
            {
               drivers.back().device = device; 
               drivers.back().tunerconf.set_string(LIBTUNER_DOMAIN_KEY, basename(device));
            }
            break;
         }
         case 'c':
            try
            {
               channel = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store channel name.";
               error = ENOMEM;
            }
            break;
         case 'u':
         {
            cx88_capture *cap = cx88_capture_factory::create(optarg, drivers.back().tunerconf, error);
            if (cap == NULL)
            {
               break;
            }
            try
            {
               drivers.back().captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store URL capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 't':
            error = drivers.back().tunerconf.load_file(optarg);
            if (error)
            {
               CX88ERR << "Unable to load libtuner configuration file '" << optarg << "': " << strerror(error) << endl; 
            }
            break;
         case 'n':
            duration_s = (unsigned long) (strtod(optarg, NULL) * 60);
            break;
         case 'o':
            drivers.back().tunerconf.load_string(optarg, ';');
            break;
         case 'p':
            remote_port = (unsigned short) (strtoul(optarg, NULL, 10));
            remote_enable = true;
            break;
         case 'l':
            params.lock_timeout_ms = (uint32_t) (strtoul(optarg, NULL, 10));
            break;
         case 'q':
            interactive = false;
            break;
         case 'h':
            auto_start = false;
            break;
         case 'a':
            switch_audio = true;
            break;
         case 'e':
            list_devices(xmlconf, "Digital video devices:", "/dev/cx88mpeg*");
            list_devices(xmlconf, "Analog video devices:", "/dev/cx88video*");
            list_devices(xmlconf, "Analog audio devices:", "/dev/cx88audio*");
            failquiet = true;
            break;
         case 'b':
            drivers.back().capture_mode = cx88_driver::CX88_CAPTURE_MODE_BACKEND;
            break;
         default:
            break;
      }
   }
   if (!error && (drivers.front().device == NULL))
   {
      if (!failquiet)
      {
         CX88ERR << "Error: You must specify a device (-d); Available devices:" << endl;
         list_devices(xmlconf, "Digital video:", "/dev/cx88mpeg*");
         list_devices(xmlconf, "Analog video:", "/dev/cx88video*");
         list_devices(xmlconf, "Analog audio:", "/dev/cx88audio*");
      }
      error = EINVAL;
   }
   cx88_driver *next = NULL;
   for (driverlist::reverse_iterator iter = drivers.rbegin(); !error && (iter != drivers.rend()); ++iter)
   {
      if (strstr(iter->device, "cx88mpeg") != NULL)
      {
         iter->driver = cx88_mpeg_driver_factory::create(
            xmlconf,
            iter->tunerconf,
            iter->device,
            iter->captures,
            error);
         if (error == ENOENT) list_devices(xmlconf, "Available digital video devices:", "/dev/cx88mpeg*");
         else if (!error && (iter->driver == NULL))
         {
            CX88ERR << "Unable to create cx88 MPEG driver" << endl;
            error = ENOMEM;  
         }
      }
      else if (strstr(iter->device, "cx88video") != NULL)
      {
         cx88_analog_driver *pip_driver = NULL;
         const char *pip_device = iter->tunerconf.get_string("pip");
         if (pip_device != NULL)
         {
            error = push_driver_entry(drivers);
            if (!error)
            {
               cx88_driver_entry &entry = drivers.back();
               entry.device = strdup(pip_device);
               if (entry.device != NULL)
               {
                  pip_driver = cx88_analog_driver_factory::create(
                     xmlconf,
                     entry.tunerconf,
                     entry.device,
                     entry.captures,
                     cx88_analog_driver::CX88_ANALOG_MODE_VIDEO,
                     NULL,
                     NULL,
                     error);
               }
               entry.driver = pip_driver;
            }
         }
         iter->driver = cx88_analog_driver_factory::create(
            xmlconf,
            iter->tunerconf,
            iter->device,
            iter->captures,
            cx88_analog_driver::CX88_ANALOG_MODE_VIDEO,
            dynamic_cast <cx88_analog_driver*>(next),
            pip_driver,
            error);
         if (error == ENOENT) list_devices(xmlconf, "Available analog video devices:", "/dev/cx88video*");
         else if (!error && (iter->driver == NULL))
         {
            CX88ERR << "Unable to create cx88 video driver" << endl;
            error = ENOMEM;  
         }
      }
      else if (strstr(iter->device, "cx88audio") != NULL)
      {
         iter->driver = cx88_analog_driver_factory::create(
            xmlconf,
            iter->tunerconf,
            iter->device,
            iter->captures,
            (((switch_audio && (iter->captures.size() > 0)) || (!switch_audio && (iter->captures.size() == 0))) ? 
               cx88_analog_driver::CX88_ANALOG_MODE_AUDIO_OSS :
               cx88_analog_driver::CX88_ANALOG_MODE_AUDIO),
            dynamic_cast <cx88_analog_driver*>(next),
            NULL,
            error);
         if (error == ENOENT) list_devices(xmlconf, "Available analog audio devices:", "/dev/cx88audio*");
         else if (!error && (iter->driver == NULL))
         {
            CX88ERR << "Unable to create cx88 audio driver" << endl;
            error = ENOMEM;  
         }
      }
      else
      {
         CX88ERR << "Invalid device filename " << iter->device << "; Available devices:" << endl;
         list_devices(xmlconf, "Digital video:", "/dev/cx88mpeg*");
         list_devices(xmlconf, "Analog video:", "/dev/cx88video*");
         list_devices(xmlconf, "Analog audio:", "/dev/cx88audio*");
         error = EINVAL;
      }
      for (capvec::size_type i = 0; !error && (i < iter->captures.size()); ++i)
      {
         error = iter->captures[i]->set_driver(iter->driver);
      }
      next = iter->driver;
   }
   if (!error)
   {
      error = capture(drivers.front(), xmlconf, channel, duration_s, remote_port, remote_enable, auto_start, interactive, params);
   }
   for (driverlist::iterator iter = drivers.begin(); iter != drivers.end(); ++iter)
   {
      for (capvec::size_type i = 0; i < iter->captures.size(); ++i)
      {
         delete iter->captures[i];
      }
      delete iter->driver;
      free(iter->device);
   }
   if (g_controlfds[0] != -1)
   {
      close(g_controlfds[0]);
   }
   if (g_controlfds[1] != -1)
   {
      close(g_controlfds[1]);
   }
   return (failquiet ? 0 : error);
}

static void capture_complete(void *arg, int)
{
   int fd = *(reinterpret_cast <int*> (arg));
   static const char c[] = "q";
   write(fd, c, sizeof(c));
}

static void print_channel(void *arg, int error)
{
   cx88_scan_arg * scan_arg = reinterpret_cast<cx88_scan_arg *>(arg);
   sleep(1);
   double strength;
   if (!error && scan_arg->driver->has_signal(strength) &&
       ((strength == 0.0) || (strength >= scan_arg->threshold)))
   {
      ostringstream stream;
      stream << scan_arg->driver->get_channel().get_channel() << " ";
      if (strength != 0.0)
      {
         stream << "(" << strength << "%) ";
      }
      cout << stream.str().c_str();
   }
   else
   {
      cout << ". ";  
   }
   if (scan_arg->driver->get_channel().last_channel())
   {
      scan_arg->driver->stop();
   }
   else
   {
      string channel = CX88_XML_CHANNEL_UP;
      if (scan_arg->driver->set_channel(channel, scan_arg->lock_timeout_ms) != 0)
      {
         scan_arg->driver->stop();
      }
   }
   cout.flush();
}

static int create_server(unsigned short port)
{
   int server_sock;
   struct sockaddr_in server_addr;
   memset(&server_addr, 0, sizeof(server_addr));
   server_addr.sin_family = AF_INET;
   server_addr.sin_port = htons(port);
   server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   if ((server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
   {
      CX88ERR << "Unable to create remote control socket: " << strerror(errno) << endl;
      return -1;
   }
   int reuseaddr = 1;
   setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(reuseaddr));
   if (bind(server_sock, (struct sockaddr*) (&server_addr), sizeof(server_addr)) < 0) 
   {
      close(server_sock);
      CX88ERR << "Unable to bind remote control socket: " << strerror(errno) << endl;
      return -1;
   }
   if (listen(server_sock, 1) < 0)
   {
      close(server_sock);
      CX88ERR << "Unable to establish socket listener: " << strerror(errno) << endl;
      return -1;
   }
   return server_sock;
}

static int set_timer(unsigned long duration_s)
{
   signal(SIGALRM, term_handler);
   struct itimerval timerval;
   memset(&timerval, 0, sizeof(timerval));
   timerval.it_value.tv_sec = duration_s;
   if (setitimer(ITIMER_REAL, &timerval, NULL) != 0)
   {
      CX88ERR << "Unable to set capture timer: " << strerror(errno) << endl;
      return errno;
   }
   return 0;
}

static int capture(
   cx88_driver_entry &entry,
   cx88_xml_config &xmlconf,
   string &channel,
   unsigned long duration_s,
   unsigned short remote_port,
   bool remote_enable,
   bool auto_start,
   bool interactive,
   cx88_capture_params &params)
{
   cx88_driver *driver = entry.driver;
   cx88_capture *recording = NULL;
   int error = 0;
   fd_set fds;
   int server_sock = -1;
   bool halted = !auto_start;
   int readyfds = 0;
   int client_socks[MAX_TCP_CONNECTIONS];
   stdio_filebuf<char> client_bufs[MAX_TCP_CONNECTIONS];
   streambuf *errbuf = cerr.rdbuf();
   streambuf *outbuf = cout.rdbuf();
   streambuf *logbuf = clog.rdbuf();
   for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
   {
      client_socks[i] = -1;
   }
   time_t deadline = 0;
   bool keyboard_input = interactive;
   if (interactive)
   {
      cout << "Welcome to the cx88 capture app.  Enter '?' for a list of commands." << endl;
   }
   if ((error = set_timer(duration_s)) != 0)
   {
      goto cleanup;
   }
   if (remote_enable)
   {
      if ((server_sock = create_server(remote_port)) == -1)
      {
         error = errno;
         goto cleanup;
      }
   }
   driver->set_capture_mode(entry.capture_mode);
   if ((error = driver->set_channel(channel, params.lock_timeout_ms)))
   {
      CX88ERR << "Unable to set initial channel" << endl;
      goto cleanup;
   }
   if (duration_s != 0)
   {
      deadline = time(NULL) + duration_s;
   };
   if (!halted)
   {
      if ((error = driver->start(params)))
      {
         CX88ERR << "Unable to start driver" << endl;
         goto cleanup;
      }
   }
   do
   {
      if (keyboard_input)
      {
         cout << "> ";
         cout.flush();
      }
      if (readyfds <= 0)
      {
         int nfds = 0;
         FD_ZERO(&fds);
         if (interactive)
         {
            FD_SET(fileno(stdin), &fds);
            nfds = fileno(stdin) + 1;  
         }
         if (server_sock != -1)
         {
            if (server_sock >= nfds)
            {
               nfds = server_sock + 1;
            }
            FD_SET(server_sock, &fds);
         }
         for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
         {
            if (client_socks[i] != -1)
            {
               FD_SET(client_socks[i], &fds);
               if (client_socks[i] >= nfds)
               {
                  nfds = client_socks[i] + 1;
               }
            }
         }
         if (g_controlfds[0] >= nfds)
         {
            nfds = g_controlfds[0] + 1;
         }
         FD_SET(g_controlfds[0], &fds);
         if ((readyfds = select(nfds, &fds, NULL, NULL, NULL)) <= 0)
         {
            continue;
         }
      }
      --readyfds;
      int remote_socket = -1;
      keyboard_input = false;
      string line;
      vector<string> argv;
      try
      {
         if (FD_ISSET(g_controlfds[0], &fds))
         {
            FD_CLR(g_controlfds[0], &fds);
            char c = '\0';
            do
            {
               read(g_controlfds[0], &c, sizeof(c));
               if (c != '\0')
               {
                  line += c;
               }
            } while (c != '\0');
         }
         else if (FD_ISSET(fileno(stdin), &fds))
         {
            FD_CLR(fileno(stdin), &fds);
            getline(cin, line);
            if (cin.rdstate() != cin.goodbit)
            {
               interactive = false;
               continue;
            }
            keyboard_input = true;
         }
         else if ((server_sock != -1) && (FD_ISSET(server_sock, &fds)))
         {
            FD_CLR(server_sock, &fds);
            struct sockaddr_in client_addr;
            socklen_t client_addr_len;
            int client_sock = accept(server_sock, (struct sockaddr *)(&client_addr), &client_addr_len);
            if (client_sock == -1)
            {
               continue;
            }
            size_t i;
            for (i = 0; i < MAX_TCP_CONNECTIONS; ++i)
            {
               if (client_socks[i] == -1)
               {
                  client_socks[i] = client_sock;
                  new (&(client_bufs[i])) stdio_filebuf<char> (client_sock, ios_base::out);
                  break;
               }
            }
            if (i == MAX_TCP_CONNECTIONS)
            {
               close(client_sock);
               continue;
            }
            continue;
         }
         else
         {
            for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
            {
               if ((client_socks[i] != -1) && (FD_ISSET(client_socks[i], &fds)))
               {
                  FD_CLR(client_socks[i], &fds);
                  char c;
                  ssize_t count = 0;
                  do
                  {
                     count = recv(client_socks[i], &c, sizeof(c), 0);
                     if ((c != '\0') && (c != '\n'))
                     {
                        line += c;
                     }
                  } while ((count > 0) && (c != '\0') && (c != '\n'));
                  if (count <= 0)
                  {
                     client_bufs[i].~stdio_filebuf();
                     client_socks[i] = -1;
                     continue;
                  }
                  cout.rdbuf(&(client_bufs[i]));
                  clog.rdbuf(&(client_bufs[i]));
                  cerr.rdbuf(&(client_bufs[i]));
                  remote_socket = i;
                  break;
               }
            }
         }
         if (remote_socket == -1)
         {
            cout.rdbuf(outbuf);
            clog.rdbuf(logbuf);
            cerr.rdbuf(errbuf);
         }
         size_t token_start = 0;
         size_t token_end;
         do
         {
            token_end = line.find_first_of(" \t", token_start, 2);
            if (token_end != token_start)
            {
               size_t count = token_end;
               if (token_end != string::npos)
               {
                  count = token_end - token_start;
               }
               argv.push_back(line.substr(token_start, count));
            }
            token_start = token_end + 1;
         } while (token_end != string::npos);
         if ((remote_socket == -1) && (argv.size() == 1))
         {
            if (argv[0] == "c")
            {
               cout << "Enter channel:" << endl;
               getline(cin, line);
               argv.push_back(line);
            }
            else if (argv[0] == "+")
            {
               cout << "Enter capture URL:" << endl;
               getline(cin, line);
               argv.push_back(line);   
            }
            else if (argv[0] == "-")
            {
               cout << "Enter capture number:" << endl;
               getline(cin, line);
               argv.push_back(line);   
            }
         }
      }
      catch (...)
      {
         CX88ERR << "Unable to allocate memory for command parsing" << endl;
         continue;
      }
      if (argv.size() == 0)
      {
         continue;
      }
      int cmd_error = 0;
      if (argv[0] == "c")
      {
         uint32_t lock_timeout_ms = params.lock_timeout_ms; 
         if (argv.size() < 2)
         {
            CX88ERR << "Channel name or number required\n" << endl;
            continue;
         }
         if (argv.size() > 2)
         {
            lock_timeout_ms = strtoul(argv[2].c_str(), NULL, 10);
         }
         if ((cmd_error = driver->set_channel(argv[1], lock_timeout_ms)) != 0)
         {
            CX88ERR << "Unable to change channel: " << strerror(cmd_error) << endl;
         }
      }
      else if (argv[0] == "i")
      {
         try
         {
            string str;
            if (driver->stopped())
            {
               str = "Capture stopped.\n";
            }
            if (driver->get_capture_mode() == cx88_driver::CX88_CAPTURE_MODE_BACKEND)
            {
               str += "Backend-only capture mode.\n";
            }
            if ((cmd_error = driver->get_info(str)) != 0)
            {
               CX88ERR << "Unable to get driver info: " << strerror(cmd_error) << endl;  
            }
            else
            {
               str += "Device:\n\t";
               str += entry.device;
               const char *driver_name = driver->get_name();
               if (driver_name != NULL)
               {
                  str += " (";
                  str += driver_name;
                  str += ")";
               }
               str += "\n";
               const cx88_source *source = driver->get_selected_source();
               if ((source != NULL) && (source->name != NULL))
               {
                  str += "Input source:\n\t";
                  str += source->name;
                  str += "\n";
               }
               capvec &captures = driver->get_captures();
               if (captures.size() > 0)
               {
                  ostringstream strstream;
                  strstream << "Captures:" << endl;
                  for (capvec::size_type i = 0; i < captures.size(); ++i)
                  {
                     strstream << "\t" << i << "\t" << captures[i]->get_id() << endl;
                  }
                  strstream << endl;
                  str += strstream.str();
               }
               cout << str.c_str() << endl;
            }
         }
         catch (...)
         {
            CX88ERR << "Insufficient memory for info string" << endl;
            continue;
         }
      }
      else if (argv[0] == "s")
      {
         if (halted)
         {
            CX88ERR << "Unable to scan: driver is halted" << endl;
            continue;  
         }
         cx88_scan_arg arg;
         arg.driver = driver;
         arg.threshold = 50.0;
         arg.lock_timeout_ms = params.lock_timeout_ms;
         if (argv.size() >= 2)
         {
            stringstream stream(argv[1]);
            stream >> arg.threshold;
         }
         if (argv.size() > 2)
         {
            arg.lock_timeout_ms = strtoul(argv[2].c_str(), NULL, 10);
         }
         scan(arg, params);
      }
      else if (argv[0] == "h")
      {
         if (!halted)
         {
            driver->stop();
            halted = true;
            CX88LOG << "Capture stopped." << endl;
         }
         else
         {
            if ((cmd_error = driver->start(params)))
            {
               CX88ERR << "Unable to restart driver: " << strerror(cmd_error) << endl;
            }
            halted = false;
            CX88LOG << "Capture started." << endl;
         }
      }
      else if ((argv[0] == "+") || ((argv[0] == "r") && (recording == NULL)))
      {
         const char *url;
         bool is_recording = false;
         if (argv[0] == "r")
         {
            url = xmlconf.get_recording_url();
            if (url == NULL)
            {
               CX88ERR << "Recording URL not configured" << endl;
               continue;
            }
            is_recording = true;
         }
         else if (argv.size() < 2)
         {
            CX88ERR << "Capture URL required" << endl;
            continue;
         }
         else
         {
            url = argv[1].c_str();
         }
         cx88_capture *cap = cx88_capture_factory::create(url, entry.tunerconf, cmd_error);
         if ((cap == NULL) || (cmd_error != 0) || (cap->set_driver(driver) != 0) || (driver->add_capture(*cap) != 0))
         {
            delete cap;
            CX88ERR << "Unable to create capture object" << endl;
            continue;
         }
         if (is_recording)
         {
            CX88LOG << "Recording to " << url << endl;
            recording = cap;
         }
      }
      else if (argv[0] == "-")
      {
         if (argv.size() < 2)
         {
            CX88ERR << "Capture number required" << endl;
            continue;
         }
         cx88_capture *cap = driver->remove_capture(strtoul(argv[1].c_str(), NULL, 10));
         if (cap == NULL)
         {
            CX88ERR << "Capture " << argv[1] << " not found" << endl;
         }
         if (cap == recording)
         {
            recording = NULL;
         }
         delete cap;
      }
      else if (argv[0] == "r")
      {
         driver->remove_capture(recording);
         delete recording;
         recording = NULL;
         CX88LOG << "Stopped recording to " << xmlconf.get_recording_url() << endl;
      }
      else if (argv[0] == "n")
      {
         if (argv.size() == 1)
         {
            const cx88_source *sources = driver->get_sources();
            if ((sources == NULL) || (sources->name == NULL))
            {
               cout << "No non-default input sources found" << endl;
               continue;
            }
            cout << "Available input sources:\n"
                 << "-----------------------------------------\n";
            size_t i = 0;
            do
            {
               if ((sources + i) == driver->get_selected_source())
               {
                  cout << "* ";
               }
               else
               {
                  cout << "  ";
               }
               cout << setw(28) << setfill(' ') << left << sources[i].name << setw(0) << sources[i].descr << endl;
            } while (sources[++i].name != NULL);
            cout << endl;
         }
         else if (driver->set_source(argv[1]))
         {
            CX88ERR << "Unable to change input source" << endl;
         }
      }
      else if (argv[0] == "?")
      {
         cout << "Interactive commands:\n"
              << "--------------------------------------------\n"
              << "c <channel> [<timeout>]       Change channel, optionally specifying lock timeout in ms\n"
              << "s [<strength%> [<timeout>]]   Scan for channels with at least the specified signal strength\n"
              << "i                             Get current tuning information\n"
              << "h                             Toggle start/stop of all captures\n"
              << "+ <URL>                       Add a capture using the specified URL\n"
              << "- <number>                    Remove capture by index (use 'i' to get indexed capture list)\n"
              << "r                             Toggle start/stop of recording to default recording file\n"
              << "n [<source>]                  Change the active input source, or list available sources\n"
              << "p [<port>]                    Set remote control port (0 => auto), or display list of remote connections\n"
              << "p-                            Disable remote control\n"
              << "t [<minutes>]                 Set capture timer (0 => disable), or display remaining time\n"
              << "ver                           Display version\n"
              << "q                             Quit app(interactive), or close session(remote)\n"
              << "q!                            Quit app(both interactive and remote)\n";
         driver->print_help();
         cout << "\nCapture URLs:\n"
              << "--------------------------------------------\n";
         cx88_xml_config::capture_desc desc;
         while (xmlconf.next_capture_desc(desc))
         {
            if (desc.url_prefix != NULL)
            {
               cout << desc.url_prefix << setw(30 - strlen(desc.url_prefix)) << setfill(' ') << left << "://";
               if (desc.desc != NULL)
               {
                  cout << setw(0) << desc.desc;
               }
               cout << endl;
            }
         }
         cout << endl;
      }
      else if (argv[0] == "p")
      {
         if (argv.size() > 1)
         {
            unsigned short new_port = (unsigned short) (strtoul(argv[1].c_str(), NULL, 10));
            int new_server = -1;
            if ((new_server = create_server(new_port)) == -1)
            {
               continue; 
            }
            if (server_sock != -1)
            {
               close(server_sock);
            }
            server_sock = new_server;
            remote_port = new_port;
            remote_enable = true;
         }
         if (!remote_enable)
         {
            cout << "Remote connections are disabled" << endl;
            continue;
         }
         struct sockaddr_in sockinfo;
         socklen_t socksize = sizeof(sockinfo);
         memset(&sockinfo, 0, sizeof(sockinfo));
         getsockname(server_sock, (struct sockaddr*)(&sockinfo), &socksize);
         cout << "Listening for remote connections on port " << ntohs(sockinfo.sin_port) << endl << "Connected clients:" << endl;
         for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
         {
            if (client_socks[i] != -1)
            {
               char peer_str[INET_ADDRSTRLEN];
               memset(&sockinfo, 0, sizeof(sockinfo));
               if ((getpeername(client_socks[i], (struct sockaddr*)(&sockinfo), &socksize) != 0) ||
                   (inet_ntop(AF_INET, &(sockinfo.sin_addr), peer_str, sizeof(peer_str)) == NULL))
               {
                  cout << "<Address not available>" << endl;
               }
               else
               {
                  cout << peer_str << endl;
               }
            }
         }
         cout << endl;
      }
      else if (argv[0] == "p-")
      {
         if (server_sock != -1)
         {
            close(server_sock);
            server_sock = -1;
         }
         cout << "Remote connections are disabled" << endl;
         remote_enable = false;
      }
      else if (argv[0] == "t")
      {
         if (argv.size() > 1)
         {
            unsigned long new_duration = (unsigned long) (strtod(argv[1].c_str(), NULL) * 60);
            if (set_timer(new_duration) != 0)
            {
               continue;
            }
            duration_s = new_duration;
            deadline = time(NULL) + new_duration;
         }
         if (duration_s == 0)
         {
            cout << "Timer is disabled" << endl;
         }
         else
         {
            cout << "Time remaining: " << (deadline - time(NULL)) << "s" << endl; 
         }
         cout << endl;
      }
      else if (argv[0] == "ver")
      {
         cout << "cx88 " CX88_VERSION << endl;
      }
      else if (argv[0] == "q")
      {
         if ((duration_s != 0) && (time(NULL) >= deadline))
         {
            cout << "Timer expired; exiting..." << endl;
         }
         if (remote_socket != -1)
         {
            cout << "Closing remote session; use 'q!' to close the entire app remotely" << endl;
            client_bufs[remote_socket].~stdio_filebuf();
            client_socks[remote_socket] = -1;
         }
         else
         {
            break;
         }
      }
      else if (argv[0] == "q!")
      {
         break;
      }
      else
      {
         driver->command(argv);
      }
   } while (!error);
   driver->stop();

cleanup:
   for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
   {
      if (client_socks[i] != -1)
      {
         client_bufs[i].~stdio_filebuf();
      }
   }
   if (server_sock != -1)
   {
      close(server_sock);
   }
   return error;
}

int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params)
{
   int error = 0;
   string orig_channel;
   arg.driver->get_channel().get_channel_name(orig_channel);
   arg.driver->stop();
   int controlfds[2];
   if (pipe(controlfds))
   {
      CX88ERR << "Unable to create control file descriptor for scanning: " 
         << strerror(errno) << endl;  
      return errno;
   }
   cx88_capture_params scan_params;
   scan_params.completion = capture_complete;
   scan_params.completion_arg = &(controlfds[1]);
   scan_params.channel_event = print_channel;
   scan_params.channel_arg = &arg;
   scan_params.lock_timeout_ms = orig_params.lock_timeout_ms;
   fd_set fds;
   FD_ZERO(&fds);
   FD_SET(controlfds[0], &fds);
   cout << "Scanning profile " << arg.driver->get_channel().get_profile_name() << ' ';
   string start_channel = CX88_XML_CHANNEL_FIRST;
   ofstream temperr("/dev/null");
   streambuf *origerr = cerr.rdbuf(temperr.rdbuf());
   error = arg.driver->set_channel(start_channel, arg.lock_timeout_ms);
   error = (error ? error : arg.driver->start(scan_params));
   if (error)
   {
      CX88ERR << "Unable to start driver for scan" << endl;
      close(controlfds[0]);
      close(controlfds[1]);
      return error;
   }
   char cmd = '\0';
   do
   {
      read(controlfds[0], &cmd, sizeof(cmd));
   } while (cmd != 'q');
   cout << "Finished scan." << endl;
   cerr.rdbuf(origerr);
   close(controlfds[0]);
   close(controlfds[1]);
   error = arg.driver->set_channel(orig_channel, orig_params.lock_timeout_ms);
   error = (error ? error : arg.driver->start(orig_params));
   if (error)
   {
      CX88ERR << "Unable to restart driver" << endl;
      return error;
   }
   return error;
}
