/*-
 * Copyright 2013 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <assert.h>
#include <strings.h>
#include <sys/errno.h>
#include <map>
#include "cx88_debug.h"
#include "cx88_capture_factory.h"

struct cx88_capture_entry
{
   const char *name;
   cx88_capture_factory::build_capture_func builder;
};

struct cx88_url_entry
{
   cx88_url_entry() : cap(NULL) {}
   const cx88_capture_entry *cap;
   tuner_config options;
};

#define CX88_MAX_CAPTURES 16

static cx88_capture_entry captures[CX88_MAX_CAPTURES];

static uint32_t num_captures = 0;

cx88_url_entry default_handler;

struct cx88_url_key
{
   bool prefix_only;
   std::string url;
};

struct cx88_compare_urls
{
   bool operator() (const cx88_url_key &key1, const cx88_url_key &key2)
   {
      size_t len = (key1.prefix_only ? key1.url.length() : key2.url.length());
      return (strncasecmp(key1.url.c_str(), key2.url.c_str(), len) < 0);
   }
};

typedef std::map<cx88_url_key, cx88_url_entry, cx88_compare_urls> url_map;

static url_map urls;

void cx88_capture_factory::registerCapture(const char *name,
                                           cx88_capture_factory::build_capture_func builder)
{
   assert(num_captures < CX88_MAX_CAPTURES);
   captures[num_captures].name = name;
   captures[num_captures].builder = builder;
   ++num_captures;
}

void cx88_capture_factory::init(cx88_xml_config &conf, int &error)
{
   if (error)
   {
      return;
   }
   cx88_xml_config::capture_desc desc;
   while (!error && conf.next_capture_desc(desc))
   {
      if (desc.classname == NULL)
      {
         continue;
      }
      uint32_t i;
      for (i = 0; i < num_captures; ++i)
      {
         if (strcasecmp(captures[i].name, desc.classname) != 0)
         {
            continue;
         }
         if (desc.url_prefix == NULL)
         {
            if (default_handler.cap != NULL)
            {
               CX88ERR << "Warning: Duplicate default capture entry " << desc.classname << " will be ignored" << std::endl;
               break;
            }
            default_handler.cap = &captures[i];
            if (desc.options != NULL)
            {
               error = default_handler.options.load_string(desc.options, ';');
            }
            break;
         }
         try
         {
            cx88_url_key key;
            key.prefix_only = true;
            key.url = desc.url_prefix;
            key.url += "://";
            cx88_url_entry entry;
            entry.cap = &captures[i];
            if (desc.options != NULL)
            {
               error = entry.options.load_string(desc.options, ';');
            }
            if (!urls.insert(std::pair<cx88_url_key, cx88_url_entry>(key, entry)).second)
            {
               CX88ERR << "Warning: Duplicate capture entry for URL type " << key.url.c_str() << " will be ignored" << std::endl; 
            }
            break;
         }
         catch (...)
         {
            error = ENOMEM;
            return;
         }
      }
      if (i == num_captures)
      {
         CX88ERR << "Warning: Capture class " << desc.classname << " not found" << std::endl;
      }
   }
}

cx88_capture *cx88_capture_factory::create(const char *url, tuner_config &options, int &error)
{
   if (error)
   {
      return NULL;
   }
   cx88_url_key key;
   try
   {
      key.prefix_only = false;
      key.url = url;
   }
   catch (...)
   {
      error = ENOMEM;
      return NULL;
   }
   cx88_url_entry *entry = NULL;
   const char *url_name = strstr(url, "://");
   if (url_name == NULL)
   {
      entry = &default_handler;
      url_name = url;
   }
   else
   {
      url_name += 3;
      url_map::iterator iter = urls.find(key);
      entry = ((iter == urls.end()) ? NULL : &(iter->second));
   }
   if ((entry == NULL) || (entry->cap == NULL))
   {
      CX88ERR << "No URL handler found for " << url << std::endl;
      error = ENOENT;
      return NULL;
   }
   error = options.add_config(entry->options);
   cx88_capture *cap = entry->cap->builder(key.url, url_name, options, error);
   options.remove_config(entry->options);
   if (error)
   {
      delete cap;
      return NULL;
   }
   return cap;
}

