/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <sys/errno.h>
#include <new>
#include "cx88_debug.h"
#include "cx88_driver.h"

using namespace std;

const cx88_source cx88_driver::m_default_source = {NULL, NULL, cx88_driver::CX88_SOURCE_DEFAULT};

cx88_driver::cx88_driver(
   const char *name,
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : m_name(name),
     m_channel(cx88conf, tunerconf, error),
     m_tunerconf(tunerconf),
     m_devnode(devnode),
     m_stop(true),
     m_channel_changed(false),
     m_fd(-1),
     m_current_source(&m_default_source), 
     m_source(&m_default_source),
     m_explicit_source(false),
     m_captures(captures),
     m_lock_timeout_ms(1000),
     m_capture_mode(CX88_CAPTURE_MODE_NORMAL),
     m_state(CX88_DRIVER_UNINITIALIZED)
{
   i2c_init();
   pthread_mutexattr_t attrs;
   if ((error = pthread_mutexattr_init(&attrs)))
   {
      CX88ERR << "Unable to initialize lock attributes" << endl;
      return;
   }
   if ((error = pthread_mutexattr_settype(&attrs, PTHREAD_MUTEX_RECURSIVE)))
   {
      CX88ERR << "Unable to set lock type to recursive" << endl;
      goto cleanup;
   }
   if ((error = pthread_mutex_init(&m_lock, &attrs)))
   {
      CX88ERR << "Unable to initialize driver lock" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_LOCK_CREATED;
   if ((error = pthread_cond_init(&m_signal, NULL)))
   {
      CX88ERR << "Unable to initialize driver event signal" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_SIGNAL_CREATED;
   if ((error = pthread_mutex_init(&m_capture_lock, &attrs)))
   {
      CX88ERR << "Unable to initialize capture lock" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_CAPTURE_LOCK_CREATED;
   m_params.completion = NULL;
   m_params.completion_arg = NULL;
   m_params.channel_event = NULL;
   m_params.channel_arg = NULL;
   m_params.lock_timeout_ms = m_lock_timeout_ms;
cleanup:
   pthread_mutexattr_destroy(&attrs);
}

cx88_driver::~cx88_driver(void)
{
   if (m_state >= CX88_DRIVER_CAPTURE_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_capture_lock);
   }
   if (m_state >= CX88_DRIVER_SIGNAL_CREATED)
   {
      pthread_cond_destroy(&m_signal);
   }
   if (m_state >= CX88_DRIVER_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_lock);
   }
   m_state = CX88_DRIVER_UNINITIALIZED;
}

int cx88_driver::get_card_desc(cx88_xml_config &xmlconf, const char *devnode, cx88_xml_config::card_desc &desc)
{
   int fd = open(devnode, O_RDONLY);
   if (fd < 0)
   {
      CX88ERR << "Unable to open device " << devnode << ": " << strerror(errno) << endl;
      return errno;
   }
   uint32_t id = 0;
   if (ioctl(fd, CX88_COMMON_CARD_ID, &id))
   {
      CX88ERR << "Unable to issue CARD_ID ioctl: " << strerror(errno) << endl;
      close(fd);
      return errno;
   }
   close(fd);
   desc = xmlconf.get_card_desc(id);
   return 0;
}

int cx88_driver::open_bridge(void)
{
   char bridge[CX88_COMMON_MAX_DRIVER_NAME + 5];
   sprintf(bridge, "/dev/");
   int error = ioctl(m_fd, CX23885_COMMON_BRIDGE, bridge + 5);
   if (error)
   {
      CX88ERR << "Unable to retrieve bridge device: " << strerror(errno) << endl;
      return -1;
   }
   int bridge_fd = open(bridge, O_RDWR);
   if (bridge_fd < 0)
   {
      CX88ERR << "Unable to open bridge device " << bridge << ": " << strerror(errno) << endl;
      return -1;
   }
   return bridge_fd;
}

int cx88_driver::change_channel(uint32_t lock_timeout_ms)
{
   m_lock_timeout_ms = lock_timeout_ms;
   if (!m_stop)
   {
      m_channel_changed = true;
      pthread_cond_signal(&m_signal);
   }
   return 0;
}

const cx88_source *cx88_driver::find_source(const char *source_name)
{
   const cx88_source *sources = get_sources();
   size_t i = 0;
   while (sources[i].name != NULL)
   {
      if (strcasecmp(source_name, sources[i].name) == 0)
      {
         return sources + i;
      }
      ++i;
   }
   return NULL;
}

void cx88_driver::update_source(const char *source_name)
{
   const cx88_source *source = NULL;
   if (source_name != NULL)
   {
      source = find_source(source_name);
   }
   if (source != NULL)
   {
      m_source = source;
      m_explicit_source = false;
   }
   else if (!m_explicit_source)
   {
      m_source = get_sources();
   }
}

int cx88_driver::set_source(const std::string &source_name)
{
   const cx88_source *source = find_source(source_name.c_str());
   if (source == NULL)
   {
      CX88ERR << "Input source " << source_name << " not found" << endl;
      return EINVAL;
   }
   cx88_driver_lock lock(this);
   m_source = source;
   m_explicit_source = true;
   return change_channel(m_lock_timeout_ms);
}

int cx88_driver::start(cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   int error = 0;
   if (m_stop)
   {
      m_current_source = get_sources();
      if (m_source == &m_default_source)
      {
         m_source = get_sources();
      }
      m_params = params;
      m_stop = false;
      if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
      {
         CX88ERR << "Unable to create driver thread: " << strerror(error) << endl;
         m_stop = true;
         return error;
      }
   }
   return error;
}

int cx88_driver::stop(void)
{
   void *retval = NULL;
   {
      cx88_driver_lock lock(this);
      if (!m_stop)
      {
         m_stop = true;
         pthread_cond_signal(&m_signal);
      }
   }
   pthread_join(m_thread, &retval);
   return (reinterpret_cast <size_t> (retval));
}

void *cx88_driver::thread_routine(void *arg)
{
   cx88_driver *driver = reinterpret_cast<cx88_driver*>(arg);
   int error = driver->capture(driver->m_params);
   return (reinterpret_cast<void*>(error));
}

int cx88_driver::start_captures(void)
{
   cx88_capture_lock lock(this);
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->start()))
      {
         CX88ERR << "Error starting captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

int cx88_driver::add_capture(cx88_capture &cap)
{
   int error = 0;
   if (!m_stop)
   {
      error = cap.start();
   }
   if (!error)
   {
      try
      {
         cx88_capture_lock lock(this);
         m_captures.push_back(&cap);
      }
      catch (...)
      {
         cap.stop();
         error = ENOMEM;
      }
   }
   return error;
}

cx88_capture *cx88_driver::remove_capture(unsigned int index)
{
   cx88_capture *cap = NULL;
   {
      cx88_capture_lock lock(this);
      if (index >= m_captures.size())
      {
         return NULL;
      }
      cap = m_captures[index];
      m_captures.erase(m_captures.begin() + index);
   }
   cap->stop();
   return cap;
}

void cx88_driver::remove_capture(cx88_capture *cap)
{
   bool capture_found = false;
   {
      cx88_capture_lock lock(this);
      for (capvec::size_type i = 0; i < m_captures.size(); ++i)
      {
         if (m_captures[i] == cap)
         {
            m_captures.erase(m_captures.begin() + i);
            capture_found = true;
            break;
         }
      }
   }
   if (capture_found)
   {
      cap->stop();
   }
}

int cx88_driver::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   cx88_capture_lock lock(this);
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->submit_buffer(buffer, num_bytes, packet_size)))
      {
         CX88ERR << "Error submitting buffer to captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

void cx88_driver::notify_captures(uint32_t event, int error)
{
   cx88_capture_lock lock(this);
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->notify(event, error);
   }
}

void cx88_driver::stop_captures(void)
{
   cx88_capture_lock lock(this);
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->stop();
   }
}

void cx88_driver::wait_for_event(void)
{
   cx88_driver_lock lock(this);
   while (!m_stop && !m_channel_changed)
   {
      pthread_cond_wait(&m_signal, &m_lock);
   }
}

void cx88_driver::i2c_init(void)
{
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      sprintf(m_i2c_devs[i].conf.in_driver, "iicdev");
      m_i2c_devs[i].conf.in_speed = CX88_I2C_SPEED_100KHZ;
      m_i2c_devs[i].conf.in_addr = 0;
      m_i2c_devs[i].dev = NULL;
      m_i2c_devs[i].driver = NULL;
      m_i2c_devs[i].bus = 0;
      m_i2c_devs[i].source_mask = CX88_SOURCE_DEFAULT;
   }
}

int cx88_driver::i2c_setup(void)
{
   cx88_driver_lock lock(this);
   i2c_init();
   if (m_capture_mode == CX88_CAPTURE_MODE_BACKEND) return 0;
   int error = create_i2c_devices();
   for (uint8_t i = 0; i < CX23885_I2C_BUSES; ++i)
   {
      cx88_common_i2c_config config;
      config.in_bus_number = i;
      config.in_num_i2c_devices = 0;
      uint8_t config_indexes[CX88_COMMON_MAX_I2C_DEVICES];
      for (uint8_t j = 0; j < MAX_I2C_DEVS; ++j)
      {
         if ((m_i2c_devs[j].conf.in_addr != 0) && (m_i2c_devs[j].bus == i))
         {  
            config.devices[config.in_num_i2c_devices] = m_i2c_devs[j].conf;
            config_indexes[(config.in_num_i2c_devices)++] = j;
            if (config.in_num_i2c_devices == CX88_COMMON_MAX_I2C_DEVICES)
            {
               break;
            }
         }
      }
      if (config.in_num_i2c_devices == 0)
      {
         continue;
      }
      error = ioctl(m_fd, CX88_COMMON_I2C_CONFIG, &config);
      if (error)
      {
         CX88ERR << "Error issuing I2C CONFIG ioctl: " << strerror(errno) << endl;
         return errno;
      }
      for (uint8_t j = 0; j < config.in_num_i2c_devices; ++j)
      {
         char i2c_devnode[CX88_COMMON_MAX_DRIVER_NAME + 5];
         snprintf(i2c_devnode, CX88_COMMON_MAX_DRIVER_NAME, "/dev/%s%d", 
            m_i2c_devs[config_indexes[j]].conf.in_driver, config.devices[j].out_instance);
         m_i2c_devs[config_indexes[j]].dev = new(nothrow) tuner_devnode_device(m_tunerconf, i2c_devnode, error);
         if (m_i2c_devs[config_indexes[j]].dev == NULL)
         {
            error = ENOMEM;
         }
         if (error)
         {
            CX88ERR << "Unable to create control instance for peripheral device" << endl;
            i2c_teardown();
            break;
         }
      }
   }
   if (!error)
   {
      error = create_i2c_drivers();  
   }
   return error;
}

void cx88_driver::i2c_teardown(void)
{
   cx88_driver_lock lock(this);
   i2c_stop();
   destroy_i2c_drivers();
   destroy_i2c_devices();
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      if (m_i2c_devs[i].dev != NULL)
      {
         delete m_i2c_devs[i].dev;
         m_i2c_devs[i].dev = NULL;
         m_i2c_devs[i].driver = NULL;
      }
   }
}

void cx88_driver::i2c_stop(void)
{
   cx88_driver_lock lock(this);
   for (uint8_t i = MAX_I2C_DEVS; i > 0; --i)
   {
      if ((m_i2c_devs[i - 1].driver != NULL) && (m_i2c_devs[i].source_mask & m_current_source->mask))
      {   
         m_i2c_devs[i - 1].driver->stop();
      }
   }
}

int cx88_driver::i2c_start(void)
{
   cx88_driver_lock lock(this);
   int error = 0;
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {
         if (m_i2c_devs[i].source_mask & m_source->mask)
         {
            if (!error && (error = m_i2c_devs[i].driver->start(m_lock_timeout_ms)))
            {
               CX88ERR << "Error starting peripheral device" << endl;
            }
         }
         else
         {
            m_i2c_devs[i].driver->reset();
         }
      }
   }
   m_current_source = m_source;
   return 0;
}

void cx88_driver::print_help(void)
{
   cout << "b                             Toggle between backend and normal capture modes\n";
}

int cx88_driver::command(vector<string> argv)
{
   if (argv[0] == "b")
   {
      bool restart = !m_stop;
      stop();
      if (m_capture_mode == CX88_CAPTURE_MODE_BACKEND)
      {
         m_capture_mode = CX88_CAPTURE_MODE_NORMAL;
         CX88LOG << "Normal capture mode." << endl;
      }
      else
      {
         m_capture_mode = CX88_CAPTURE_MODE_BACKEND;
         CX88LOG << "Backend capture mode." << endl;
      }
      if (restart) return start();
   }
   else if (argv[0].length() > 0)
   {
      CX88ERR << "Unrecognized command: " << argv[0].c_str() << endl;
   }
   return 0;
}
