/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_DRIVER_H__
#define __CX88_DRIVER_H__

#include <sys/ioctl.h>
#include <pthread.h>
#include <string>
#include "cx88_xml_config.h"
#include "tuner_config.h"
#include "tuner_devnode_device.h"
#include "tuner_driver.h"
#include "cx88_common_ioctl.h"
#include "cx88_capture.h"

typedef void (*cx88_event)(void*, int);

typedef struct
{
   cx88_event completion;
   void *completion_arg;
   cx88_event channel_event;
   void *channel_arg;
   uint32_t lock_timeout_ms;
} cx88_capture_params;

typedef struct
{
   const char *name;
   const char *descr;
   uint32_t mask;
} cx88_source;

class cx88_driver
{
   public:

      enum cx88_capture_mode
      {
         CX88_CAPTURE_MODE_BACKEND,
         CX88_CAPTURE_MODE_NORMAL
      };

      cx88_driver(
         const char *name,
         cx88_xml_config &cx88conf,
         tuner_config &tunerconf,
         const char *devnode,
         capvec &captures,
         int &error);

      virtual ~cx88_driver(void);

      virtual int start(cx88_capture_params &params);

      virtual int start(void) { return start(m_params); }

      virtual int stop(void);

      bool stopped(void) { return m_stop; }

      virtual int set_channel(const std::string &channel_string, uint32_t lock_timeout_ms) = 0;

      virtual int set_source(const std::string &source_name);

      virtual int get_info(std::string &info) = 0;

      virtual bool has_signal(double &strength) = 0;
      
      cx88_xml_channel &get_channel(void)
      {
         return m_channel;
      }
      
      virtual int command(std::vector<std::string> argv);

      virtual void print_help(void);

      virtual int add_capture(cx88_capture &cap);

      virtual cx88_capture* remove_capture(unsigned int index);

      virtual void remove_capture(cx88_capture *cap);

      virtual capvec &get_captures(void)
      {
         return m_captures;
      }

      virtual const char *get_name(void)
      {
         return m_name;
      }

      virtual uint32_t get_caps(void)
      {
         return 0;
      }

      virtual const cx88_source *get_sources(void)
      {
         return &m_default_source;
      }

      const cx88_source *get_selected_source(void)
      {
         return m_source;
      }

      static const uint32_t CX88_EVENT_CHANNEL = (1 << 0);

      static int get_card_desc(cx88_xml_config &xmlconf, const char *devnode, cx88_xml_config::card_desc &desc);

      int open_bridge(void);

      void set_capture_mode(cx88_capture_mode mode)
      {
         m_capture_mode = mode;
      }

      cx88_capture_mode get_capture_mode(void)
      {
         return m_capture_mode;
      }

   protected:

      static const uint8_t MAX_I2C_DEVS = CX23885_I2C_BUSES * CX88_COMMON_MAX_I2C_DEVICES;
      struct cx88_i2c_device
      {
         tuner_devnode_device *dev;
         tuner_driver *driver;
         cx88_i2c_device_config conf;
         uint32_t source_mask;
         uint8_t bus;
      };
      cx88_i2c_device m_i2c_devs[MAX_I2C_DEVS];
      const char *m_name;
      cx88_xml_channel m_channel;
      tuner_config &m_tunerconf;
      const char *m_devnode;
      volatile bool m_stop;
      volatile bool m_channel_changed;
      int m_fd;
      const cx88_source *m_current_source;
      const cx88_source *m_source;
      bool m_explicit_source;
      capvec &m_captures;
      uint32_t m_lock_timeout_ms;
      cx88_capture_params m_params;
      cx88_capture_mode m_capture_mode;

      int start_captures(void);      
      int submit_buffer(void *buffer, size_t num_bytes, size_t packet_size = 0);
      void notify_captures(uint32_t event, int error);
      void stop_captures(void);

      void wait_for_event(void);
      
      virtual int capture(cx88_capture_params &params) = 0;

      int change_channel(uint32_t lock_timeout_ms);

      const cx88_source *find_source(const char *source_name);

      void update_source(const char *source_name);

      void i2c_init(void);
      
      virtual int i2c_setup(void);
      
      virtual void i2c_teardown(void);
      
      virtual int create_i2c_devices(void)
      {
         return 0;
      }
      
      virtual int create_i2c_drivers(void) 
      {
         return 0;
      }
      
      virtual void destroy_i2c_drivers(void) {}
      
      virtual void destroy_i2c_devices(void) {}

      static const cx88_source m_default_source;

      static const uint32_t CX88_SOURCE_DEFAULT = 1 << 0;
      
      virtual int select_source(uint32_t source_mask)
      {
         return 0;
      }
      
      void i2c_stop(void);
      int i2c_start(void);
            
   private:
      
      friend class cx88_driver_lock;
      friend class cx88_capture_lock;
    
      pthread_t m_thread;
      pthread_mutex_t m_lock;
      pthread_cond_t m_signal;
      pthread_mutex_t m_capture_lock;
      enum
      {
         CX88_DRIVER_UNINITIALIZED,
         CX88_DRIVER_LOCK_CREATED,
         CX88_DRIVER_SIGNAL_CREATED,
         CX88_DRIVER_CAPTURE_LOCK_CREATED
      } m_state;

      static void *thread_routine(void *arg);
};

class cx88_driver_lock
{
   public:
      
      cx88_driver_lock(cx88_driver *driver)
         : m_driver(driver)
      {
         pthread_mutex_lock(&(m_driver->m_lock));
      }
   
      virtual ~cx88_driver_lock(void)
      {
         pthread_mutex_unlock(&(m_driver->m_lock));
      }
      
   private:
      
      cx88_driver *m_driver;
};

class cx88_capture_lock
{
   public:
      
      cx88_capture_lock(cx88_driver *driver)
         : m_driver(driver)
      {
         pthread_mutex_lock(&(m_driver->m_capture_lock));
      }
   
      virtual ~cx88_capture_lock(void)
      {
         pthread_mutex_unlock(&(m_driver->m_capture_lock));
      }
      
   private:
      
      cx88_driver *m_driver;
};

#endif
