/*
 * Copyright 2013 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include "cx88_debug.h"
#include "cx23888_dif.h"

class tuner_dummy_device : public tuner_device
{
   public:

      tuner_dummy_device(tuner_config &tunerconf) : tuner_device(tunerconf) {}

      virtual ~tuner_dummy_device() {}

      virtual int write(const uint8_t*, size_t, size_t&) { return 0; }

      virtual int read(uint8_t*, size_t, size_t&) { return 0; }

};

static tuner_config dummy_config;
static tuner_dummy_device dummy_device(dummy_config);

cx23888_dif::cx23888_dif(cx88_driver &driver, tuner_config &tunerconf, int &error)
   : tuner_driver(tunerconf, dummy_device),
     avb_driver(tunerconf, dummy_device),
     m_fd(-1)
{
   m_fd = driver.open_bridge();
   if (m_fd == -1)
   {
      CX88ERR << "Error opening CX23888 bridge: " << strerror(errno) << std::endl;
      error = ENOENT;
   }
}

cx23888_dif::~cx23888_dif()
{
   if (m_fd != -1)
   {
      reset();
      close(m_fd);
      m_fd = -1;
   }
}

int cx23888_dif::set_channel(const avb_channel &channel)
{
   int error = 0;
   cx23888_dif_control dif_control;
   static const uint8_t fm_dif_config[] = {
      0x3d, 0xa,  0xd7, 0x9d, 0xc,  0xbc, 0x3,  0x65,
      0x85, 0x8c, 0x3,  0xbd, 0xa,  0x64, 0xb4, 0x1d,
      0x0,  0x88, 0x0,  0x0,  0x0,  0x18, 0x40, 0x22,
      0x0,  0x26, 0x0,  0x40, 0x0,  0x26, 0x26, 0xc2,
      0x0,  0x26, 0x26, 0xc2, 0x0,  0x26, 0x0,  0x40,
      0x0,  0xc2, 0x0,  0x0,  0x20, 0xe6, 0x50, 0x2,
      0x0,  0x1,  0x0,  0x27, 0x13, 0x2b, 0x31, 0x0,
      0x0,  0x0,  0x0,  0x0,  0x0,  0xc8, 0x0,  0x68,
      0x6,  0xbf, 0xef, 0x1b, 0xe8, 0x35, 0x0,  0x0,
      0xa,  0x0,  0x1,  0x0,  0x61, 0x0,  0xb,  0x0,
      0xad, 0x0,  0x57, 0x0,  0xf4, 0x1,  0xe1, 0x0,
      0xca, 0x1,  0xeb, 0x2,  0x64, 0x7,  0x9a, 0x0,
      0x1b, 0x5,  0x35, 0x6,  0x60, 0x3,  0x24, 0x4,
      0xac, 0x12, 0xdd, 0x12, 0x69, 0x13, 0xd7, 0x12,
      0xdb, 0x35, 0x6a, 0x14, 0x6,  0x3a, 0xbc, 0x37,
      0xa8, 0x3f, 0xaf, 0x3c, 0x38, 0x6,  0xdd, 0x2,
      0xf7, 0xc,  0x9f, 0x9,  0xb,  0x13, 0x24, 0x10,
      0xa4, 0x17, 0x93, 0x15, 0x1d, 0x1a, 0x2c, 0x19,
      0x0,  0x0,  0x6f, 0x1a, 0x0,  0x0,  0x0,  0x0
   };
   switch (channel.audio_format)
   {
      case AVB_AUDIO_FMT_FM_MONO:
      case AVB_AUDIO_FMT_FM_MONO_NON_USA:
      case AVB_AUDIO_FMT_FM_MONO_USA:
      case AVB_AUDIO_FMT_FM_STEREO:
      case AVB_AUDIO_FMT_FM_STEREO_NON_USA:
      case AVB_AUDIO_FMT_FM_STEREO_USA:
         dif_control.regs = fm_dif_config;
         dif_control.offset = 0;
         dif_control.count = sizeof(fm_dif_config);
         break;
      default:
         CX88ERR << "Audio format " << channel.audio_format <<
            "not currently supported for CX3888 demodulator" << std::endl;
         error = EINVAL;
         break;
   }
   if (!error && (ioctl(m_fd, CX23888_COMMON_DIF_ENABLE, &dif_control) == -1))
   {
      CX88ERR << "Error configuring CX23888 demodulator: " << strerror(errno) << std::endl;
      error = errno;
   }
   return error;
}

void cx23888_dif::reset(void)
{
   if (ioctl(m_fd, CX23888_COMMON_DIF_DISABLE) == -1)
   {
      CX88ERR << "Error disabling CX23888 demodulator: " << strerror(errno) << std::endl;
   }
}

