/*-
 * Copyright 2009 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <new>
#include <string>
#include "xc5000.h"
#include "cx88_mpeg_driver.h"
#include "s5h1411.h"
#include "cx88_debug.h"
#include "cx88_mpeg_driver_factory.h"

namespace cx88_mpeg_driver_factory
{

   class fusion_7dual
      : public cx88_mpeg_driver
   {
      public:
         
         fusion_7dual(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            int &error);
         
         virtual ~fusion_7dual(void) {}

         virtual const char *get_name(void);

         virtual uint32_t get_caps(void)
         {
            return CX88_MPEG_CAP_ATSC | CX88_MPEG_CAP_QAM;
         }
         
      protected:
         
         int peripheral_reset(uint8_t mask);

         static int xc5000_reset(xc5000 &driver, void *context);

         virtual int create_i2c_drivers(void);

         virtual void destroy_i2c_drivers(void);

         virtual int create_i2c_devices(void);
    
         uint8_t m_bus;

         std::string m_name;

   };

   CX88_REGISTER_MPEG_DRIVER(fusion_7dual)

   fusion_7dual::fusion_7dual(
      const char *name,
      cx88_xml_config &cx88conf,
      tuner_config &tunerconf,
      const char *devnode,
      capvec &captures,
      int &error)
      : cx88_mpeg_driver(name, cx88conf, tunerconf, devnode, captures, error),
        m_bus(0)
   {
      int fd = open(devnode, O_RDONLY);
      uint32_t stream = 0;
      if ((fd < 0) || (ioctl(fd, CX23885_COMMON_GET_STREAMS, &stream) != 0))
      {
         CX88ERR << "Error retrieving stream type: " << strerror(errno) << std::endl;
         error = errno;
      }
      if (stream == CX23885_STREAM_MPEG2)
      {
         m_bus = 1;
      }
      if (fd >= 0)
      {
         close(fd);
      }
      try
      {
         m_name = name;
         m_name += ((m_bus == 0) ? " tuner #1" : " tuner #2");
      }
      catch (...)
      {
         CX88ERR << "Unable to allocate name buffer" << std::endl;
         error = (error ? error : ENOMEM);
      }
   }

   const char *fusion_7dual::get_name(void)
   {
      return m_name.c_str();
   }

   int fusion_7dual::peripheral_reset(uint8_t mask)
   {
      cx88_common_gpio gpio;
      gpio.in_write_mask = mask;
      gpio.in_read_mask = 0x0;
      gpio.inout_gpio = 0x0;
      if (ioctl(m_fd, CX88_COMMON_GPIO, &gpio) != 0)
      {
         CX88ERR << "Error during GPIO ioctl: " << strerror(errno) << std::endl;
         return errno;
      }
      usleep(20000);
      gpio.inout_gpio = mask;
      if (ioctl(m_fd, CX88_COMMON_GPIO, &gpio) != 0)
      {
         CX88ERR << "Error during GPIO ioctl: " << strerror(errno) << std::endl;
         return errno;
      }
      usleep(20000);
      return 0;
   }

   int fusion_7dual::xc5000_reset(xc5000 &driver, void *context)
   {
      fusion_7dual *arg = reinterpret_cast<fusion_7dual*>(context);
      return arg->peripheral_reset((arg->m_bus == 1) ? 0x4 : 0x1);
   }

   int fusion_7dual::create_i2c_drivers(void)
   {
      int error = peripheral_reset((m_bus == 1) ? 0x8 : 0x2);
      m_i2c_devs[1].driver = new(std::nothrow) s5h1411(m_tunerconf, *(m_i2c_devs[1].dev), *(m_i2c_devs[2].dev), DVB_INPUT_SERIAL, 
         S5H1411_IFREQ_5_38MHZ, S5H1411_IFREQ_5_38MHZ, S5H1411_GPIO_ENABLE, S5H1411_CLK_CONT_NINV, error);
      m_i2c_devs[0].driver = new(std::nothrow) xc5000(m_tunerconf, *(m_i2c_devs[0].dev), 5380000, xc5000_reset, this, error);
      if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
      {
         error = ENOMEM;
      }
      if (error)
      {
         CX88ERR << "Error initializing tuner/demodulator driver for Fusion 7 Dual" << std::endl;
         return error;
      }
      return 0;
   }

   void fusion_7dual::destroy_i2c_drivers(void)
   {
      delete m_i2c_devs[0].driver;
      delete m_i2c_devs[1].driver;  
   }

   int fusion_7dual::create_i2c_devices(void)
   {
      m_i2c_devs[0].bus = m_bus;
      m_i2c_devs[0].conf.in_addr = 0x64;
      m_i2c_devs[0].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      m_i2c_devs[1].bus = m_bus;
      m_i2c_devs[1].conf.in_addr = 0x19;
      m_i2c_devs[1].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      m_i2c_devs[2].bus = m_bus;
      m_i2c_devs[2].conf.in_addr = 0x1A;
      m_i2c_devs[2].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      return 0;
   }
}
