/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/errno.h>
#include <new>
#include "cx88_mpeg_driver.h"
#include "cx88_analog_driver.h"
#include "lgh064f.h"
#include "lg3303.h"
#include "tda9887.h"
#include "cx88_debug.h"
#include "cx88_mpeg_driver_factory.h"
#include "cx88_analog_driver_factory.h"

namespace cx88_mpeg_driver_factory
{

   class pchdtv_hd5500
      : public cx88_mpeg_driver
   {
      public:
         
         pchdtv_hd5500(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            int &error)
            : cx88_mpeg_driver(name, cx88conf, tunerconf, devnode, captures, error)
         {}
         
         virtual ~pchdtv_hd5500(void) {}

         virtual uint32_t get_caps(void)
         {
            return CX88_MPEG_CAP_ATSC | CX88_MPEG_CAP_QAM;
         }

      protected:
         
         virtual int create_i2c_drivers(void);

         virtual void destroy_i2c_drivers(void);

         virtual int create_i2c_devices(void);

         virtual int set_capture_options(cx88_mpeg_capture &capture);
   };

   CX88_REGISTER_MPEG_DRIVER(pchdtv_hd5500)
}

namespace cx88_analog_driver_factory
{
   class pchdtv_hd5500
      : public cx88_analog_driver
   {
      public:
         
         pchdtv_hd5500(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            cx88_analog_mode mode,
            cx88_analog_driver *slave,
            cx88_analog_driver *pip_driver,
            int &error)
            : cx88_analog_driver(name, cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error)
         {}
         
         virtual ~pchdtv_hd5500(void) {}

         virtual uint32_t get_caps(void)
         {
            return CX88_ANALOG_CAP_TV | CX88_ANALOG_CAP_SVIDEO | CX88_ANALOG_CAP_COMPOSITE |
                   CX88_ANALOG_CAP_NTSC | CX88_ANALOG_CAP_TV_AUDIO | CX88_ANALOG_CAP_FM | CX88_ANALOG_CAP_MUTE;
         }

         virtual const cx88_source *get_sources(void)
         {
            return m_sources;
         }

      protected:

         static const uint32_t HD5500_SOURCE_SVIDEO = 1 << 1;
         static const uint32_t HD5500_SOURCE_COMPOSITE = 1 << 2;
         static const cx88_source m_sources[4];

         uint8_t m_input_mux;
         uint8_t m_input_type;
         
         virtual int create_i2c_drivers(void);

         virtual void destroy_i2c_drivers(void);

         virtual int create_i2c_devices(void);

         virtual int select_source(uint32_t source_mask);

         virtual int set_capture_options(cx88_video_capture &capture);

   };

   CX88_REGISTER_ANALOG_DRIVER(pchdtv_hd5500)
};

namespace cx88_mpeg_driver_factory
{
   int pchdtv_hd5500::create_i2c_drivers(void)
   {
      int error = 0;
      cx88_common_gpio gpio;
      gpio.in_write_mask = 0x1;
      gpio.in_read_mask = 0x0;
      gpio.inout_gpio = 0x0;
      if ((error = ioctl(m_fd, CX88_COMMON_GPIO, &gpio)) != 0)
      {
         CX88ERR << "Error issuing GPIO ioctl: " << strerror(errno) << std::endl;
         return error;
      }
      usleep(100000);
      gpio.in_write_mask = 0x1;
      gpio.in_read_mask = 0x0;
      gpio.inout_gpio = 0x1;
      if ((error = (ioctl(m_fd, CX88_COMMON_GPIO, &gpio)) != 0))
      {
         CX88ERR << "Error issuing GPIO ioctl: " << strerror(errno) << std::endl;
         return error;  
      }
      usleep(200000);
      m_i2c_devs[0].driver = new(std::nothrow) lgh064f(m_tunerconf, *(m_i2c_devs[0].dev));
      m_i2c_devs[1].driver = new(std::nothrow) lg3303(m_tunerconf, *(m_i2c_devs[1].dev), DVB_IFC_NEG_POL, DVB_INPUT_SERIAL, error);
      if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
      {
         CX88ERR << "Error initializing tuner/demodulator driver for pcHDTV HD-5500" << std::endl;
         return ENOMEM;
      }
      return error;
   }
   
   void pchdtv_hd5500::destroy_i2c_drivers(void)
   {
      delete m_i2c_devs[1].driver;
      delete m_i2c_devs[0].driver;  
   }
   
   int pchdtv_hd5500::create_i2c_devices(void)
   {
      m_i2c_devs[0].conf.in_addr = 0x61;
      m_i2c_devs[1].conf.in_addr = 0x59;
      return 0;
   }
   
   int pchdtv_hd5500::set_capture_options(cx88_mpeg_capture &capture)
   {
      capture.in_options = CX88_MPEG_SERIAL_INPUT | CX88_MPEG_BYTEWIDE_SOP;
      return 0;
   }

}

namespace cx88_analog_driver_factory
{
   const cx88_source pchdtv_hd5500::m_sources[] =
      {{"Tuner",     "RF input",                  cx88_driver::CX88_SOURCE_DEFAULT},
       {"SVideo",    "S-video input",             pchdtv_hd5500::HD5500_SOURCE_SVIDEO},
       {"Composite", "Composite video input",     pchdtv_hd5500::HD5500_SOURCE_COMPOSITE},
       {NULL,        NULL,                        0}};

   int pchdtv_hd5500::create_i2c_drivers(void)
   {
      m_i2c_devs[0].driver = new(std::nothrow) lgh064f(m_tunerconf, *(m_i2c_devs[0].dev));
      m_i2c_devs[1].driver = new(std::nothrow) tda9887(m_tunerconf, *(m_i2c_devs[1].dev), TDA9887_OPTION_RADIO_IF_41_3 | TDA9887_OPTION_RADIO_GAIN_NORM);
      if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
      {
         CX88ERR << "Error initializing analog tuner/demodulator drivers for pcHDTV HD-5500" << std::endl;
         return ENOMEM;
      }
      return 0;
   }

   void pchdtv_hd5500::destroy_i2c_drivers(void)
   {
      delete m_i2c_devs[0].driver;
      delete m_i2c_devs[1].driver;
   }

   int pchdtv_hd5500::create_i2c_devices(void)
   {
      m_i2c_devs[0].conf.in_addr = 0x61;
      m_i2c_devs[0].source_mask = CX88_SOURCE_DEFAULT;
      m_i2c_devs[1].conf.in_addr = 0x43;
      m_i2c_devs[1].source_mask = CX88_SOURCE_DEFAULT;
      return 0;
   }

   int pchdtv_hd5500::select_source(uint32_t source_mask)
   {
      switch (source_mask)
      {
         case CX88_SOURCE_DEFAULT:
            m_input_mux = CX88_VIDEO_INPUT_MUX1;
            m_input_type = CX88_VIDEO_INPUT_COMPOSITE;
            break;
         case HD5500_SOURCE_SVIDEO:
            m_input_mux = CX88_VIDEO_INPUT_MUX3;
            m_input_type = CX88_VIDEO_INPUT_SVIDEO;
            break;
         case HD5500_SOURCE_COMPOSITE:
            m_input_mux = CX88_VIDEO_INPUT_MUX2;
            m_input_type = CX88_VIDEO_INPUT_COMPOSITE;
            break;
         default:
            return EINVAL;
      }
      return 0;
   }

   int pchdtv_hd5500::set_capture_options(cx88_video_capture &capture)
   {
      capture.in_input_source = m_input_mux;
      capture.in_input_type = m_input_type;
      return 0;
   }
}

