/*-
 * Copyright 2012 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <new>
#include "xc3028.h"
#include "cx88_analog_driver.h"
#include "cx88_debug.h"
#include "cx88_analog_driver_factory.h"

namespace cx88_analog_driver_factory
{

   class videomate_e800
      : public cx88_analog_driver
   {
      public:
         
         videomate_e800(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            cx88_analog_mode mode,
            cx88_analog_driver *slave,
            cx88_analog_driver *pip_driver,
            int &error)
            : cx88_analog_driver(name, cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error)
         {}
         
         virtual ~videomate_e800(void) {}

         virtual uint32_t get_caps(void)
         {
            return CX88_ANALOG_CAP_FM | CX88_ANALOG_CAP_VOLUME | CX88_ANALOG_CAP_MUTE;
         }

      protected:

         virtual int create_i2c_drivers(void);

         virtual void destroy_i2c_drivers(void);

         virtual int create_i2c_devices(void);

         static int xc3028_reset(xc3028::xc3028_reset_t reset_type, void *context);
   };

   CX88_REGISTER_ANALOG_DRIVER(videomate_e800)
}

static int peripheral_reset(int fd, uint16_t mask, xc3028::xc3028_reset_t reset_type)
{
   if (reset_type != xc3028::XC3028_TUNER_RESET)
   {
      return 0;
   }
   cx88_common_gpio gpio;
   gpio.in_write_mask = mask;
   gpio.in_read_mask = 0x0;
   gpio.inout_gpio = 0x0;
   if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
   {
      CX88ERR << "Error during GPIO ioctl: " << strerror(errno) << std::endl;
      return errno;
   }
   usleep(20000);
   gpio.inout_gpio = mask;
   if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
   {
      CX88ERR << "Error during GPIO ioctl: " << strerror(errno) << std::endl;
      return errno;
   }
   usleep(20000);
   return 0;
}

namespace cx88_analog_driver_factory
{
   int videomate_e800::create_i2c_drivers(void)
   {
      int error = peripheral_reset(m_fd, 0x4, xc3028::XC3028_TUNER_RESET);
      xc3028 *tuner;
      m_i2c_devs[0].driver = tuner = new(std::nothrow) xc3028(m_tunerconf, *(m_i2c_devs[0].dev), xc3028_reset, this, error);
      if (m_i2c_devs[0].driver == NULL)
      {
         error = (error ? error : ENOMEM);
      }
      if (error)
      {
         CX88ERR << "Error initializing tuner driver for VideoMate EXXX" << std::endl;
         return error;
      }
      //tuner->set_firmware_flags(XC3028_BASEFW_INPUT1, 0, 0, 0, 0, 0); 
      return 0;
   }

   void videomate_e800::destroy_i2c_drivers(void)
   {
      delete m_i2c_devs[0].driver;
   }

   int videomate_e800::create_i2c_devices(void)
   {
      m_i2c_devs[0].bus = 1;
      m_i2c_devs[0].conf.in_addr = 0x61;
      m_i2c_devs[0].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      return 0;
   }

   int videomate_e800::xc3028_reset(xc3028::xc3028_reset_t reset_type, void *context)
   {
      videomate_e800 *driver = (videomate_e800*)context;
      return peripheral_reset(driver->m_fd, 0x4, reset_type);
   }
}

