/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <new>
#include "tda18271.h"
#include "cx88_debug.h"
#include "cx88_mpeg_driver.h"
#include "cx88_analog_driver.h"
#include "s5h1411.h"
#include "cx88_mpeg_driver_factory.h"
#include "cx88_analog_driver_factory.h"
#include "cx23888_dif.h"

namespace cx88_mpeg_driver_factory
{

   class wintv_hvr1850
      : public cx88_mpeg_driver
   {
      public:
         
         wintv_hvr1850(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            int &error)
            : cx88_mpeg_driver(name, cx88conf, tunerconf, devnode, captures, error)
         {
            m_inversion_on = tunerconf.get_number<unsigned int>("HVR1850_INVERSION", 1);
         }
         
         virtual ~wintv_hvr1850(void) {}

         virtual uint32_t get_caps(void)
         {
            return CX88_MPEG_CAP_ATSC | CX88_MPEG_CAP_QAM;
         }

      protected:

         static int tda18271_interface_config(tda18271 &driver, const dvb_channel &channel, tda18271_interface &ifc);

         virtual int create_i2c_drivers(void);

         virtual void destroy_i2c_drivers(void);

         virtual int create_i2c_devices(void);
 
         virtual int set_channel_options(dvb_channel &channel);

         unsigned int m_inversion_on;

   };

   CX88_REGISTER_MPEG_DRIVER(wintv_hvr1850)
};

namespace cx88_analog_driver_factory
{
   class wintv_hvr1850
      : public cx88_analog_driver
   {
      public:
         
         wintv_hvr1850(
            const char *name,
            cx88_xml_config &cx88conf,
            tuner_config &tunerconf,
            const char *devnode,
            capvec &captures,
            cx88_analog_mode mode,
            cx88_analog_driver *slave,
            cx88_analog_driver *pip_driver,
            int &error);

         virtual ~wintv_hvr1850(void) {}

         virtual uint32_t get_caps(void);

         virtual const cx88_source *get_sources(void);

      protected:

         static const uint32_t HVR1850_SOURCE_LINEIN = 1 << 1;

         static const cx88_source m_sources[3];

         bool m_secondary_audio;
         uint8_t m_audio_input;

         virtual int select_source(uint32_t source_mask);

         virtual int set_format_options(cx88_audio_format &format);

         virtual int create_i2c_devices(void);

         virtual void destroy_i2c_drivers(void);

         static int tda18271_interface_config(tda18271 &driver, const avb_channel &channel, tda18271_interface &ifc);

         virtual int create_i2c_drivers(void);

   };

   CX88_REGISTER_ANALOG_DRIVER(wintv_hvr1850)
};

namespace cx88_mpeg_driver_factory
{

   int wintv_hvr1850::tda18271_interface_config(tda18271 &driver, const dvb_channel &channel, tda18271_interface &ifc)
   {
      ifc.std = 0x0;
      ifc.if_level = 0x06;
      switch (channel.modulation)
      {
         case DVB_MOD_VSB_8:
         case DVB_MOD_VSB_16:
            ifc.std = 0x03;
            ifc.ifreq_hz = 5380000;
            break;
         default:
            break;
      }
      return 0;
   }

   int wintv_hvr1850::create_i2c_drivers(void)
   {
      int error = 0;
      cx88_common_gpio gpio;
      gpio.in_write_mask = 0x6000;
      gpio.in_read_mask = 0x0;
      gpio.inout_gpio = 0x0;
      if (ioctl(m_fd, CX23885_COMMON_MC_GPIO, &gpio) != 0)
      {
         CX88ERR << "Error during microcontroller GPIO ioctl: " << strerror(errno) << std::endl;
         return errno;
      }
      usleep(100000);
      gpio.inout_gpio = 0x4000;
      if (ioctl(m_fd, CX23885_COMMON_MC_GPIO, &gpio) != 0)
      {
         CX88ERR << "Error during microcontroller GPIO ioctl: " << strerror(errno) << std::endl;
         return errno;
      }
      usleep(100000);
      m_i2c_devs[1].driver = new(std::nothrow) s5h1411(m_tunerconf, *(m_i2c_devs[1].dev), *(m_i2c_devs[2].dev), DVB_INPUT_SERIAL, 
         S5H1411_IFREQ_5_38MHZ, S5H1411_IFREQ_4MHZ, S5H1411_GPIO_DISABLE, S5H1411_CLK_CONT_NINV, error);
      m_i2c_devs[0].driver = new(std::nothrow) tda18271(m_tunerconf, *(m_i2c_devs[0].dev), TDA18271_MODE_SINGLE,
            NULL, tda18271_interface_config, error);
      if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
      {
         error = ENOMEM;
      }
      if (error)
      {
         CX88ERR << "Error initializing tuner/demodulator driver for WinTV HVR-1850" << std::endl;
         return error;
      }
      return 0;
   }

   void wintv_hvr1850::destroy_i2c_drivers(void)
   {
      delete m_i2c_devs[0].driver;
      delete m_i2c_devs[1].driver;  
   }
   
   int wintv_hvr1850::create_i2c_devices(void)
   {
      m_i2c_devs[0].conf.in_addr = 0x60;
      // The TDA18271 on the 1850 will NAK 400KHz transactions.
      // This is odd, since the datasheet indicates it supports
      // 400KHz, and the earlier-revision TDA18271 on
      // the 1800 seems fine w/ 400KHz.
      //m_i2c_devs[0].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      m_i2c_devs[1].conf.in_addr = 0x19;
      m_i2c_devs[1].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      m_i2c_devs[2].conf.in_addr = 0x1A;
      m_i2c_devs[2].conf.in_speed = CX88_I2C_SPEED_400KHZ;
      return 0;
   }

   int wintv_hvr1850::set_channel_options(dvb_channel &channel) 
   {
      if (m_inversion_on && ((channel.modulation == DVB_MOD_VSB_8) || (channel.inversion == DVB_INVERSION_AUTO)))
      {
         channel.inversion = DVB_INVERSION_ON;
      }
      return 0;
   }
}

namespace cx88_analog_driver_factory
{

   const cx88_source wintv_hvr1850::m_sources[] = 
      {{"Tuner",     "Analog tuner",       cx88_driver::CX88_SOURCE_DEFAULT},
       {"Line-In",   "Line-In audio jack", wintv_hvr1850::HVR1850_SOURCE_LINEIN},
       {NULL,        NULL,                 0}};

   wintv_hvr1850::wintv_hvr1850(
      const char *name,
      cx88_xml_config &cx88conf,
      tuner_config &tunerconf,
      const char *devnode,
      capvec &captures,
      cx88_analog_mode mode,
      cx88_analog_driver *slave,
      cx88_analog_driver *pip_driver,
      int &error)
      : cx88_analog_driver(name, cx88conf, tunerconf, devnode, captures, mode, slave, pip_driver, error),
        m_secondary_audio(false)
   {
      int fd = open(devnode, O_RDONLY);
      uint32_t stream = 0;
      if ((fd < 0) || (ioctl(fd, CX23885_COMMON_GET_STREAMS, &stream) != 0))
      {
         CX88ERR << "Error retrieving stream type: " << strerror(errno) << std::endl;
         error = errno;
      }
      if (stream == CX23885_STREAM_AUDIO2)
      {
         m_secondary_audio = true;
      }
      if (fd >= 0)
      {
         close(fd);
      }
   }

   uint32_t wintv_hvr1850::get_caps(void)
   {
      uint32_t caps = CX88_ANALOG_CAP_FM;
      if (!m_secondary_audio)
      {
         caps |= (CX88_ANALOG_CAP_LINE_IN | CX88_ANALOG_CAP_VOLUME | CX88_ANALOG_CAP_MUTE);
      }
      return caps;
   }

   const cx88_source *wintv_hvr1850::get_sources(void)
   {
      return (m_secondary_audio ? cx88_driver::get_sources() : m_sources);
   }

   int wintv_hvr1850::select_source(uint32_t source_mask)
   {
      m_audio_input = CX23885_AUDIO_INPUT_BCAST2;
      if (source_mask == HVR1850_SOURCE_LINEIN)
      {
         m_audio_input = CX23885_AUDIO_INPUT_AC97;
      }
      if (m_secondary_audio)
      {
         m_audio_input |= CX23885_AUDIO_INPUT_DEMOD_STEREO;
      }
      return 0;
   }

   int wintv_hvr1850::set_format_options(cx88_audio_format &format)
   {
      format.in_input = m_audio_input;
      return 0;
   }

   int wintv_hvr1850::create_i2c_devices(void)
   {
      m_i2c_devs[0].bus = 1;
      m_i2c_devs[0].conf.in_addr = 0x60;
      return 0;
   }

   void wintv_hvr1850::destroy_i2c_drivers(void)
   { 
      delete m_i2c_devs[0].driver;
      delete m_i2c_devs[1].driver;
   }

   int wintv_hvr1850::tda18271_interface_config(tda18271 &driver, const avb_channel &channel, tda18271_interface &ifc)
   {
      ifc.fm_rfn = 0;
      return 0;
   }

   int wintv_hvr1850::create_i2c_drivers(void)
   {
      int error = 0;
      m_i2c_devs[0].driver = new(std::nothrow) tda18271(m_tunerconf, *(m_i2c_devs[0].dev), TDA18271_MODE_SINGLE,
            tda18271_interface_config, NULL, error);
      m_i2c_devs[1].driver = new(std::nothrow) cx23888_dif(*this, m_tunerconf, error);
      if ((m_i2c_devs[0].driver == NULL) || (m_i2c_devs[1].driver == NULL))
      {
         error = ENOMEM;
      }
      if (error)
      {
         CX88ERR << "Error initializing analog tuner driver for WinTV HVR-1850" << std::endl;
         return error;
      }
      return error;
   }
}

