/*-
 * Copyright 2012 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include "cx88_debug.h"
#include "cx88_hal.h"

using namespace std;

cx88_hal::cx88_hal(void)
   : m_dbus_conn(NULL),
     m_hal_context(NULL),
     m_udi(NULL),
     m_initialized(false)
{
   m_dbus_conn = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
   if (m_dbus_conn == NULL)
   {
      CX88LOG << "Warning: Unable to establish DBus connection for HAL registration" << endl;
      return;
   }
   m_hal_context = libhal_ctx_new();
   if (m_hal_context == NULL)
   {
      CX88LOG << "Warning: Unable to allocate HAL context" << endl;
      return;
   }
	if (!libhal_ctx_set_dbus_connection(m_hal_context, m_dbus_conn))
   {
      CX88LOG << "Warning: Unable to set DBus connection for HAL context" << endl;
		return;
   }
   if (!libhal_ctx_init(m_hal_context, NULL))
   {
      CX88LOG << "Warning: Unable to establish HAL connection" << endl;
      return;
   }
   m_initialized = true;
}

cx88_hal::~cx88_hal(void)
{
   stop();
   if (m_hal_context != NULL)
   {
      libhal_ctx_shutdown(m_hal_context, NULL);
      libhal_ctx_free(m_hal_context);
   }
   if (m_dbus_conn != NULL)
   {
      dbus_connection_unref(m_dbus_conn);
   }
}

void cx88_hal::stop(void)
{
   if (m_udi != NULL)
   {
      libhal_remove_device(m_hal_context, m_udi, NULL);
      m_udi = NULL;
   }
}

int cx88_hal::start(const char *path, const cx88_hal_property *props, size_t num_props, const char **caps, size_t num_caps)
{
   if (!m_initialized)
   {
      return ENXIO;
   }
   if (m_udi != NULL)
   {
      return EBUSY;
   }
	if (libhal_device_exists(m_hal_context, path, NULL))
   {
      m_udi = path;
      return 0;
   }
   char *device = libhal_new_device(m_hal_context, NULL);
   if (device == NULL)
   {
      return ENOMEM;
   }
   LibHalChangeSet *changeset = libhal_device_new_changeset(device);
   if (changeset == NULL)
   {
      free(device);
      return ENOMEM;
   }
   int error = 0;
	if (!libhal_changeset_set_property_string(changeset, "info.parent", CX88_HAL_DEVICE_PATH "/computer"))
   {
      error = EINVAL;
   }
   for (size_t i = 0; !error && (i < num_props); ++i)
   {
      if (!libhal_changeset_set_property_string(changeset, props[i].name, props[i].value))
      {
         error = EINVAL;
      }
   }
   for (size_t i = 0; !error && (i < num_caps); ++i)
   {
		if (!libhal_device_add_capability(m_hal_context, device, caps[i], NULL))
      {
         error = EINVAL;
      }
   }
   if (!error && !libhal_device_commit_changeset(m_hal_context, changeset, NULL))
   {
      error = EINVAL;
   }
   libhal_device_free_changeset(changeset);
   if (!error)
   {
      m_udi = path;
   }
   if (!error && !libhal_device_commit_to_gdl(m_hal_context, device, m_udi, NULL))
   {
      error = EINVAL;
   }
   if (error)
   {
      m_udi = NULL;
   }
   free(device);
   return error;
}

