/*-
 * Copyright 2011 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <signal.h>
#include "../cx88_debug.h"
#include "cx88_capture_factory.h"
#include "cx88_radio_capture.h"

CX88_REGISTER_CAPTURE(cx88_radio_capture)

using namespace std;

const struct cuse_methods cx88_radio_capture::m_cuse_methods =
{
   radio_open,
   radio_close,
   NULL, // read
   NULL, // write
   radio_ioctl,
   NULL // poll
};

static cx88_radio_capture* cuse_owner = NULL;

static void hup_handler(int) {}

cx88_radio_capture::cx88_radio_capture(const string &id, const char *filename,
                                       tuner_config&, int &error)
   : cx88_capture(id, error),
     m_state(CX88_RADIO_CAPTURE_UNINITIALIZED),
     m_cuse_dev(NULL),
     m_driver(NULL),
     m_frequency(0),
     m_channel_status(0),
     m_opened(false)
{
   memset(&m_tuner, 0, sizeof(m_tuner));   
   m_tuner.type = V4L2_TUNER_RADIO;
   m_tuner.rangelow = 76 * 16000;
   m_tuner.rangehigh = 108 * 16000;
   m_tuner.rxsubchans = V4L2_TUNER_SUB_MONO | V4L2_TUNER_SUB_STEREO;
   m_tuner.capability = V4L2_TUNER_CAP_LOW | V4L2_TUNER_CAP_STEREO;
   memset(&m_caps, 0, sizeof(m_caps));
   strncpy((char*)(m_caps.driver), "cx88audio", sizeof(m_caps.driver) - 1);
   strncpy((char*)(m_caps.bus_info), "PCI", sizeof(m_caps.bus_info) - 1);
   m_caps.capabilities = V4L2_CAP_TUNER | V4L2_CAP_RADIO;
   memset(&m_audio, 0, sizeof(m_audio));
   strncpy((char*)(m_audio.name), "Radio", sizeof(m_audio.name) - 1);
   m_audio.capability = V4L2_AUDCAP_STEREO;
   if (error)
   {
      return;
   }
   if (!__sync_bool_compare_and_swap(&cuse_owner, NULL, this))
   {
      error = EALREADY;
      CX88ERR << "Radio capture already in progress; only one allowed per device session" << endl;
   }
   error = (error ? error : pthread_mutex_init(&m_lock, NULL));
   if (!error)
   {
      m_state = CX88_RADIO_CAPTURE_LOCK_CREATED;
   }
   error = (error ? error : pthread_cond_init(&m_event, NULL));
   if (!error)
   {
      m_state = CX88_RADIO_CAPTURE_CV_CREATED;
   }
   error = (error ? error : cuse_init());
   if (!error)
   {
      m_state = CX88_RADIO_CAPTURE_CUSE_INITIALIZED;
   }
   #define CX88_RADIO_MAX_UNIT 63
   int unit = 0;
   char node[8];
   bool auto_name = ((filename == NULL) || (strlen(filename) == 0));
   do
   {
      if (auto_name)
      {
         snprintf(node, sizeof(node), "radio%d", unit);
         filename = node;
      }
      if (!error)
      {
         m_cuse_dev = cuse_dev_create(&m_cuse_methods, this, NULL, 0, 5, 0666, "%s", filename);
      }
   } while (auto_name && (m_cuse_dev == NULL) && (unit++ < CX88_RADIO_MAX_UNIT));
   if (m_cuse_dev == NULL)
   {
      CX88ERR << "Unable to create radio device node" << endl;
      error = ENOMEM;
   }
   if (auto_name)
   {
      try
      {
         m_id += filename;
      }
      catch (...)
      {
         CX88ERR << "Unable to set radio device node name" << endl;
         error = ENOMEM;
      }
   }
   signal(SIGHUP, hup_handler);
   error = (error ? error : pthread_create(&m_cuse_thread, NULL, cuse_thread_routine, NULL));
   if (!error)
   {
      m_state = CX88_RADIO_CAPTURE_THREAD_CREATED;
   }
}

cx88_radio_capture::~cx88_radio_capture(void)
{
   if (m_cuse_dev != NULL)
   {
      cuse_dev_destroy(m_cuse_dev);
      m_cuse_dev = NULL;
   }
   if (m_state >= CX88_RADIO_CAPTURE_CUSE_INITIALIZED)
   {
      cuse_uninit();
   }
   __sync_bool_compare_and_swap(&cuse_owner, this, NULL);
   if (m_state >= CX88_RADIO_CAPTURE_THREAD_CREATED)
   {
      pthread_cancel(m_cuse_thread);
      pthread_join(m_cuse_thread, NULL);
   }
   if (m_state >= CX88_RADIO_CAPTURE_CV_CREATED)
   {
      pthread_cond_destroy(&m_event);
   }
   if (m_state >= CX88_RADIO_CAPTURE_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_lock);
   }
   m_state = CX88_RADIO_CAPTURE_UNINITIALIZED;
}

int cx88_radio_capture::set_driver(cx88_driver *driver)
{
   cx88_analog_driver *analog_driver = dynamic_cast<cx88_analog_driver*>(driver);
   if ((analog_driver == NULL) || (analog_driver->get_mode() == cx88_analog_driver::CX88_ANALOG_MODE_VIDEO) ||
       !(analog_driver->get_caps() & cx88_analog_driver::CX88_ANALOG_CAP_FM))
   {
      CX88ERR << "Inappropriate device type for radio capture: must be an analog audio device that supports FM radio" << endl;
      return EINVAL;
   }
   pthread_mutex_lock(&m_lock);
   m_driver = analog_driver;
   const char *name = m_driver->get_name();
   if (name != NULL)
   {
      strncpy((char*)(m_tuner.name), name, sizeof(m_tuner.name) - 1);
      strncpy((char*)(m_caps.card), name, sizeof(m_caps.card) - 1);
   }
   pthread_mutex_unlock(&m_lock);
   return 0;
}

void *cx88_radio_capture::cuse_thread_routine(void *arg)
{
   while (cuse_wait_and_process() == 0);
   return NULL;
}

int cx88_radio_capture::radio_open(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_radio_capture::radio_open()" << endl)
   cx88_radio_capture *cap = (cx88_radio_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   int error = 0;
   pthread_mutex_lock(&(cap->m_lock));
   if (cap->m_opened || (cap->m_driver == NULL))
   {
      error = CUSE_ERR_BUSY;
   }
   else if (cap->m_driver->start())
   {
      error = CUSE_ERR_FAULT;
   }
   else
   {
      cap->m_opened = true;
   }
   pthread_mutex_unlock(&(cap->m_lock));
   return error;
}

int cx88_radio_capture::radio_close(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_radio_capture::radio_close()" << endl)
   cx88_radio_capture *cap = (cx88_radio_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   cx88_analog_driver *driver = NULL;
   pthread_mutex_lock(&(cap->m_lock));
   if (cap->m_opened)
   {
      cap->m_opened = false;
      driver = cap->m_driver;
   }
   pthread_mutex_unlock(&(cap->m_lock));
   if (driver != NULL)
   {
      driver->stop();
   }
   return 0;
}

int cx88_radio_capture::radio_ioctl(struct cuse_dev *dev, int fflags, unsigned long cmd, void *user_data)
{
   DIAGNOSTIC(CX88LOG << "cx88_radio_capture::radio_ioctl(" << hex << cmd << ")" << endl)
   cx88_radio_capture *cap = (cx88_radio_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   return cap->ioctl(fflags, cmd, user_data);
}

void cx88_radio_capture::process_notification(uint32_t events, int error)
{
   if (events & cx88_driver::CX88_EVENT_CHANNEL)
   {
      m_channel_status = error;
      pthread_cond_signal(&m_event);
   }
}

int cx88_radio_capture::ioctl(int fflags, unsigned long code, void *user_data)
{
   int error = 0;
   pthread_mutex_lock(&m_lock);
   if (m_driver == NULL)
   {
      error = CUSE_ERR_BUSY;
   }
   else
   {
      switch (code)
      {
         case VIDIOC_QUERYCAP:
            error = cuse_copy_out(&m_caps, user_data, sizeof(m_caps)); 
            break;
         case VIDIOC_G_TUNER:
         {
            cx88_audio_status status;
            m_driver->get_status(status);
            if (status.out_signal == CX88_AUDIO_SIGNAL_STEREO)
            {
               m_tuner.audmode = V4L2_TUNER_MODE_STEREO;
            }
            else
            {
               m_tuner.audmode = V4L2_TUNER_MODE_MONO;
            }
            m_tuner.signal = ((status.out_carrier == CX88_AUDIO_NO_CARRIER) ? 0 : 0xFFFF);          
            error = cuse_copy_out(&m_tuner, user_data, sizeof(m_tuner));
            break;
         }
         case VIDIOC_S_TUNER:
         {
            struct v4l2_tuner tuner;
            if ((error = cuse_copy_in(user_data, &tuner, sizeof(tuner))))
            {
               break;
            }
            if (tuner.index != 0)
            {
               error = CUSE_ERR_INVALID;
            }
            break;
         }
         case VIDIOC_S_FREQUENCY:
         {
            struct v4l2_frequency freq;
            if ((error = cuse_copy_in(user_data, &freq, sizeof(freq))))
            {
               break;
            }
            if ((freq.tuner != 0) || (freq.type != V4L2_TUNER_RADIO))
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            avb_channel channel;
            channel.video_format = AVB_VIDEO_FMT_NONE;
            channel.audio_format = AVB_AUDIO_FMT_FM_STEREO_USA;
            channel.frequency_hz = (freq.frequency * 1000) / 16;
            channel.bandwidth_hz = 200000;
            if (m_driver->set_channel(channel))
            {
               error = CUSE_ERR_FAULT;
               break;
            }
            pthread_cond_wait(&m_event, &m_lock);
            if (m_channel_status != 0)
            {
               error = CUSE_ERR_FAULT;
            }
            else
            {
               m_frequency = freq.frequency;
            }
            break;
         }
         case VIDIOC_G_FREQUENCY:
         {
            struct v4l2_frequency freq;
            if ((error = cuse_copy_in(user_data, &freq, sizeof(freq))))
            {
               break;
            }
            if (freq.tuner != 0)
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            freq.type = V4L2_TUNER_RADIO;
            freq.frequency = m_frequency;
            error = cuse_copy_out(&freq, user_data, sizeof(freq));
            break;
         }
         case VIDIOC_G_INPUT:
         {
            unsigned int input = 0;
            error = cuse_copy_out(&input, user_data, sizeof(input));
            break;
         }
         case VIDIOC_S_INPUT:
         {
            unsigned int input;
            error = cuse_copy_in(user_data, &input, sizeof(input));
            if (!error && (input != 0))
            {
               error = CUSE_ERR_INVALID;
            }
            break;
         }
         case VIDIOC_G_AUDIO:
            error = cuse_copy_out(&m_audio, user_data, sizeof(m_audio));
            break;
         case VIDIOC_S_AUDIO:
         {
            struct v4l2_audio audio;
            error = cuse_copy_in(user_data, &audio, sizeof(audio));
            if (!error && (audio.index != 0))
            {
               error = CUSE_ERR_INVALID;
            }
            break;
         }
         case VIDIOC_QUERYCTRL:
         {
            struct v4l2_queryctrl qc;
            if ((error = cuse_copy_in(user_data, &qc, sizeof(qc))))
            {
               break;
            }
            if ((m_driver->get_caps() & cx88_analog_driver::CX88_ANALOG_CAP_MUTE) && (qc.id == V4L2_CID_AUDIO_MUTE))
            {
               qc.minimum = 0;
               qc.maximum = 1;
               qc.step = 1;
               qc.default_value = 0;
               strncpy((char*)(qc.name), "Mute", sizeof(qc.name) - 1);
            }
            else if ((m_driver->get_caps() & cx88_analog_driver::CX88_ANALOG_CAP_VOLUME) && (qc.id == V4L2_CID_AUDIO_VOLUME))
            {
               qc.minimum = 0;
               qc.maximum = 100;
               qc.step = 1;
               qc.default_value = 75;
               strncpy((char*)(qc.name), "Volume", sizeof(qc.name) - 1);
            }
            else
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            error = cuse_copy_out(&qc, user_data, sizeof(qc));
            break;
         }
         case VIDIOC_G_CTRL:
         {
            struct v4l2_control ctrl;
            if ((error = cuse_copy_in(user_data, &ctrl, sizeof(ctrl))))
            {
               break;
            }
            cx88_audio_status status;
            m_driver->get_status(status);
            if (ctrl.id == V4L2_CID_AUDIO_MUTE)
            {
               ctrl.value = status.out_mute;
            }
            else if (ctrl.id == V4L2_CID_AUDIO_VOLUME)
            {
               ctrl.value = ((status.out_vol_left > status.out_vol_right) ? status.out_vol_left : status.out_vol_right);
            }
            else
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            error = cuse_copy_out(&ctrl, user_data, sizeof(ctrl));
            break;
         }
         case VIDIOC_S_CTRL:
         {
            struct v4l2_control ctrl;
            if ((error = cuse_copy_in(user_data, &ctrl, sizeof(ctrl))))
            {
               break;
            }
            if (ctrl.id == V4L2_CID_AUDIO_MUTE)
            {
               cx88_audio_status status;
               m_driver->get_status(status);
               if ((status.out_mute != ctrl.value) && (m_driver->mute() != 0))
               {
                  error = CUSE_ERR_FAULT;
               }
            }
            else if (ctrl.id == V4L2_CID_AUDIO_VOLUME)
            {
               cx88_audio_volume vol;
               vol.in_left = ctrl.value;
               vol.in_right = ctrl.value;
               if (m_driver->set_volume(vol) != 0)
               {
                  error = CUSE_ERR_FAULT;
               }
            }
            else
            {
               error = CUSE_ERR_INVALID;
            }
            break; 
         }
         default:
            error = CUSE_ERR_INVALID;
            break;
      }
   }
   pthread_mutex_unlock(&m_lock);
   return error;
}

